from tko.play.keys import GuiActions, GuiKeys
from tko.game.xp import XP
from tko.play.opener import Opener
from tko.settings.settings import Settings
from tko.util.text import Text, Token,  RToken
from tko.util.symbols import symbols

from tko.play.fmt import Fmt
from tko.play.frame import Frame
from tko.play.border import Border
from tko.play.search import Search
from tko.play.language_setter import LanguageSetter
from tko.play.images import opening, random_get
from tko.play.floating import Floating, FloatingInput
from tko.play.floating_manager import FloatingManager
from tko.play.flags import Flag, Flags, FlagsMan
from tko.play.tasktree import TaskTree

from typing import List, Any, Dict, Callable, Tuple
import datetime

class Gui:

    def __init__(self, tree: TaskTree, flagsman: FlagsMan, fman: FloatingManager):
        self.rep = tree.rep
        self.game = tree.game
        self.tree = tree
        self.flagsman = flagsman
        self.fman = fman
        self.settings = tree.settings
        self.search = Search(tree=self.tree, fman=self.fman)
        self.style: Border = Border(self.settings.app)
        self.language = LanguageSetter(self.rep, self.flagsman, self.fman)
        self.colors = self.settings.colors

        self.app = Settings().app


    def get_help_fixed(self):
        help_fixed: List[Text] = [
            Text() + RToken("C", f" {GuiActions.leave}  [{GuiKeys.key_quit}]"),
            Text() + RToken("Y", f"{GuiActions.palette} [{GuiKeys.palette}]"),
            Text() + RToken("G", f"{GuiActions.activate} [↲]"),
            Text() + RToken("Y", f"{GuiActions.edit} [{GuiKeys.edit}]"),
            # Text() + RToken("C", f"{GuiActions.navegar} [wasd]")
            Text() + RToken("C", f"{GuiActions.search} [{GuiKeys.search}]")
        ]
        return help_fixed

    def make_flags_bar(self) -> Text:
        lista: list[Flag] = []
        lista.append(Flags.flags)
        lista.append(Flags.skills)
        lista.append(Flags.admin)
        lista.append(Flags.percent)
        lista.append(Flags.minimum)
        lista.append(Flags.reward)
        value = "1" if self.settings.app.has_borders() else "0"
        borders = Flag().set_values([value]).set_name("Bordas").set_keycode("B")
        lista.append(borders)
        value = "1" if self.settings.app.has_images() else "0"
        images = Flag().set_values([value]).set_name("Imagens").set_keycode("I")
        lista.append(images)

        values = [self.style.get_flag_sentence(flag, 0 , True, False, False) for flag in lista]
        return Text("").join(values)

    def center_header_footer(self, value: Text, frame: Frame) -> Text:
        half = value.len() // 2
        x = frame.get_x()
        dy, dx = Fmt.get_size()
        full = Text().add("─" * ((dx//2) - x - 2 - half)).add(value)
        return full

    def show_main_bar(self, frame: Frame):
        top = Text()
        alias_color = "R"
        top.add(self.style.border(alias_color, self.rep.alias.upper()))
        if Flags.admin.is_true():
            color = "W" if Flags.admin.is_true() else "K"
            top.add(self.style.border(color, "ADMIN"))
        top.add(self.style.border("G", self.rep.get_lang().upper()))
        full = self.center_header_footer(top, frame)
        frame.set_header(full, "<")
        
        if Flags.flags.is_true():
            value = self.make_flags_bar()
            full = self.center_header_footer(value, frame)
            frame.set_footer(full, "<")
        frame.draw()

        dy, dx = frame.get_inner()
        for y, sentence in enumerate(self.tree.get_senteces(dy)):
            if sentence.len() > dx:
                sentence.trim_end(dx - 3)
                sentence.addf("r", "...")
            frame.write(y, 0, sentence)



    def show_skills_bar(self, frame_xp):
        dy, dx = frame_xp.get_inner()
        xp = XP(self.game)
        total_perc = int(
            100 * (xp.get_xp_total_obtained() / xp.get_xp_total_available())
        )
        if Flags.percent.is_true():
            text = f" XPTotal:{total_perc}%"
        else:
            text = f" XPTotal:{xp.get_xp_total_obtained()}"

        done = self.colors.main_bar_done
        todo = self.colors.main_bar_todo
        total_bar = self.style.build_bar(text, total_perc / 100, dx - 2, done, todo)
        frame_xp.set_header(Text().addf("/", "Skills"), "^", "{", "}")
        frame_xp.set_footer(Text().add(" ").add(self.app._rootdir).add(" "), "^")
        frame_xp.draw()

        total, obt = self.game.get_skills_resume([self.game.quests[key] for key in self.game.available_quests])
        elements: List[Text] = []
        for skill, value in total.items():
            if Flags.percent.is_true():
                text = f"{skill}:{int(100 * obt[skill] / value)}%"
            else:
                text = f"{skill}:{obt[skill]}/{value}"
            perc = obt[skill] / value
            done = self.colors.progress_skill_done
            todo = self.colors.progress_skill_todo
            skill_bar = self.style.build_bar(text, perc, dx - 2, done, todo)
            elements.append(skill_bar)
            
        elements.append(total_bar)

        line_breaks = dy - len(elements)
        for skill_bar in elements:
            frame_xp.print(1, skill_bar)
            if line_breaks > 0:
                line_breaks -= 1
                frame_xp.print(1, Text())


    def build_list_sentence(self, items: List[Text]) -> List[Text]:
        out: List[Text] = []
        for x in items:
            color_ini = x.data[0].fmt
            color_end = x.data[-1].fmt
            left = self.style.roundL(color_ini)
            right = self.style.roundR(color_end)
            middle = x.clone()
            if x.data[0].text == "!":
                left = self.style.sharpL(color_ini)
                right = self.style.sharpR(color_end)
                middle.data = x.data[1:]
            out.append(Text().add(left).add(middle).add(right))
        return out

    def build_bottom_array(self):
        array: List[Text] = []
        array += self.get_help_fixed()
        return self.build_list_sentence(array)

    def show_bottom_bar(self):
        lines, cols = Fmt.get_size()
        elems = self.build_bottom_array()
        line_main = Text(" ").join(elems) # alignment adjust
        Fmt.write(lines - 1, 0, line_main.center(cols))

    def make_xp_button(self, size):
        if self.search.search_mode:
            text = " Busca: " + self.tree.search_text + symbols.cursor.text
            percent = 0.0
            done = "W"
            todo = "W"
            text = text.ljust(size)
        else:
            text, percent = self.build_xp_bar()
            done = self.colors.main_bar_done
            todo = self.colors.main_bar_todo
            text = text.center(size)
        xpbar = self.style.build_bar(text, percent, len(text), done, todo)
        return xpbar

    def show_top_bar(self, frame: Frame):
        lista = [
            Text() + RToken("Y", f"{GuiActions.self_grade} {GuiKeys.inc_self}{GuiKeys.dec_self}"),
            Text() + RToken("Y", f"{GuiActions.progress} {GuiKeys.dec_prog}{GuiKeys.inc_prog}"),
            # Text() + RToken("Y", f"{GuiActions.pesquisar} {GuiKeys.pesquisar}"),
        ]
        help = self.build_list_sentence(lista)
        self_grade = help[0]
        progress = help[1]
        # pesquisar = help[2]

        pre: List[Text] = []
        pre.append(progress)

        pos: List[Text] = []
        # pos.append(pesquisar)
        # skills = self.style.get_flag_sentence(Flags.skills)
        pos.append(self_grade)
        # pos.append(skills)


        limit = frame.get_dx()
        size = limit - Text(" ").join(pre + pos).len() - 2
        main_label = self.make_xp_button(size)
        info = Text(" ").join(pre + [main_label] + pos)
        frame.write(0, 0, info.center(frame.get_dx()))

    def show_help(self):
        # def empty(value):
        #     pass
        _help: Floating = Floating("v>").set_text_ljust()
        self.fman.add_input(_help)

        _help.set_header_sentence(Text().add(" Ajuda "))
        # _help.put_text(" Movimentação ".center(dx, symbols.hbar.text))
        _help.put_sentence(Text("    Ajuda ").addf("r", GuiKeys.key_help).add("  Abre essa tela de ajuda")
        )

        _help.put_sentence(Text("  ").addf("r", "Shift + B")
                           .add("  Habilita ").addf("r", "").addf("R", "ícones").addf("r", "").add(" se seu ambiente suportar"))
        _help.put_sentence(Text() + "" + RToken("g", "setas") + ", " + RToken("g", "wasd")  + "  Para navegar entre os elementos")
        _help.put_sentence(Text() + f"{GuiActions.github} " + RToken("r", f"{GuiKeys.github_open}") + "  Abre tarefa em uma aba do browser")
        _help.put_sentence(Text() + f"   {GuiActions.download} " + RToken("r", f"{GuiKeys.down_task}") + "  Baixa tarefa de código para seu dispositivo")
        _help.put_sentence(Text() + f"   {GuiActions.edit} " + RToken("r", f"{GuiKeys.edit}") + "  Abre os arquivos no editor de código")
        _help.put_sentence(Text() + f"   {GuiActions.activate} " + RToken("r", "↲") + "  Interage com o elemento")
        _help.put_sentence(Text() + f"   {GuiActions.progress} " + RToken("r", f"{GuiKeys.inc_self}") + RToken("r", f"{GuiKeys.dec_self}") + " Muda a autoavaliação")
        _help.put_sentence(Text())
        _help.put_sentence(Text() + "Você pode mudar o editor padrão com o comando")
        _help.put_sentence(Text() + RToken("g", "             tko config --editor <comando>"))


    def build_xp_bar(self) -> Tuple[str, float]:
        xp = XP(self.game)
        if xp.get_xp_total_obtained() == xp.get_xp_total_available():
            text = "Você atingiu o máximo de xp!"
            percent = 100.0
        else:
            # lang = self.rep.get_lang().upper()
            level = xp.get_level()
            percent = float(xp.get_xp_level_current()) / float(xp.get_xp_level_needed())
            if Flags.percent.is_true():
                xpobt = int(100 * xp.get_xp_level_current() / xp.get_xp_level_needed())
                text = "Level:{} XP:{}%".format(level, xpobt)
            else:
                xpobt1 = xp.get_xp_level_current()
                xpobt2 = xp.get_xp_level_needed()
                text = "Level:{} XP:{}/{}".format(level, xpobt1, xpobt2)

        return text, percent

    def show_opening(self, frame: Frame):
        # if Fmt.get_size()[1] < 100:
        #     return
        # if not self.app.has_images():
        #     return
        _, cols = Fmt.get_size()
        
        now = datetime.datetime.now()
        parrot = random_get(opening, str(now.hour))
        parrot_lines = parrot.split("\n")
        max_len = max([len(line) for line in parrot_lines])
        yinit = 1
        for y, line in enumerate(parrot_lines):
            frame.write(y, 80, Text().addf("g", line))

    def show_items(self):
        border_color = "r" if Flags.admin.is_true() else ""
        Fmt.clear()
        self.tree.reload_sentences()
        lines, cols = Fmt.get_size()
        main_sx = cols  # tamanho em x livre
        main_sy = lines  # size em y avaliable

        top_y = -1
        top_dy = 1  #quantas linhas o topo usa
        bottom_dy = 1 # quantas linhas o fundo usa
        mid_y = top_dy # onde o meio começa
        mid_sy = main_sy - top_dy - bottom_dy # tamanho do meio
        left_size = 25
        skills_sx = 0
        flags_sx = 0
        if Flags.skills.is_true():
            skills_sx = left_size #max(20, main_sx // 4)
        
        task_sx = main_sx - flags_sx - skills_sx

        frame_top = Frame(top_y, 0).set_size(top_dy + 2, cols)
        self.show_top_bar(frame_top)

        # frame_bottom = Frame(lines - bottom_dy - 1, -1).set_size(bottom_dy + 2, cols + 2)
        self.show_bottom_bar()
        if task_sx > 5: 
            frame_main = Frame(mid_y, 0).set_size(mid_sy, task_sx).set_border_color(border_color)
            self.show_main_bar(frame_main)
        self.show_opening(frame_main)

        if Flags.skills.is_true():
            frame_skills = Frame(mid_y, cols - skills_sx).set_size(mid_sy, skills_sx).set_border_color(border_color)
            self.show_skills_bar(frame_skills)



from typing import Dict, Callable, Optional
import curses
import os

class InputManager:
    backspace1 = 127
    backspace2 = 263
    left = 260
    right = 261
    a = 97
    d = 100
    tab = 9
    special_double_key = 195
    cedilha = 167
    esc = 27
    plus = 465
    minus = 464
    delete = 330

    def __init__(self):
        # stores a function than can return another function
        self.calls: Dict[int, Callable[[], Optional[Callable[[], None]]]] = {}

    def add_int(self, _key: int, fn: Callable[[], None]):
        if _key in self.calls.keys():
            raise ValueError(f"Chave duplicada {chr(_key)}")
        self.calls[_key] = fn

    def add_str(self, str_key: str, fn: Callable[[], None]):
        if str_key != "":
            self.add_int(ord(str_key), fn)

    def has_str_key(self, key: str) -> bool:
        return ord(key) in self.calls
    
    def has_int_key(self, key: int) -> bool:
        return key in self.calls
    
    def exec_call(self, key: int):
        return self.calls[key]()

    @staticmethod
    def fix_esc_delay():
        if hasattr(curses, "set_escdelay"):
            curses.set_escdelay(25)
        else:
            os.environ.setdefault('ESCDELAY', '25')

    @staticmethod
    def fix_cedilha(scr, value: int) -> int:
        if value == InputManager.special_double_key:
            value = scr.getch()
            if value == InputManager.cedilha: #ç
                value = ord("c")
        return value


import subprocess
from .text import Text, Token
from .raw_terminal import RawTerminal
from ..play.images import compilling
from .runner import Runner
from ..run.solver_builder import SolverBuilder
import random

class Free:
    @staticmethod
    def free_run(solver: SolverBuilder, show_compilling:bool=True, to_clear: bool=True, wait_input:bool=True) -> bool:

        if to_clear:
            Runner.clear_screen()
        if show_compilling:
            image = random.choice(list(compilling.keys()))
            for line in compilling[image].split("\n"):
                print(Text().addf("y", line).center(RawTerminal.get_terminal_size(), Token(" ")))

        if show_compilling:
            Runner.clear_screen()
        solver.prepare_exec(free_run_mode=True)
        if solver.compile_error:
            print(solver.error_msg)
        else:
            cmd = solver.get_executable()
            print(Text().center(RawTerminal.get_terminal_size(), Token("─")))
            answer = subprocess.run(cmd, shell=True, text=True)
            if answer.returncode != 0 and answer.returncode != 1:
                print(Runner.decode_code(answer.returncode))
        solver.reset()
        to_run_again = False
        if wait_input:
            print(Text().center(RawTerminal.get_terminal_size(), Token("─")))
            print(Text().addf("y", "Deseja compilar e executar novamente? [").addf("c", "S").addf("y", "/n]: "), end="")
            valor = input()
            if valor != "n" and valor != "q":
                if to_clear:
                    Runner.clear_screen()
                to_run_again = True
        if to_clear:
            Runner.clear_screen()

        return to_run_again


from typing import List, Optional, Tuple

from ..util.symbols import symbols
from ..util.text import Text, Token

class DiffBuilder:
    vinput    = " INSERIDO "
    vexpected = " ESPERADO "
    vreceived = " RECEBIDO "
    vunequal  = " DESIGUAL "

    def __init__(self, width: int):
        self.width = width
        self.curses = False

    def set_curses(self):
        self.curses = True
        return self

    @staticmethod
    def make_line_arrow_up(a: str, b: str) -> Text:
        hdiff = Text()
        first = True
        i = 0
        lim = max(len(a), len(b))
        while i < lim:
            if i >= len(a) or i >= len(b) or a[i] != b[i]:
                if first:
                    first = False
                    hdiff += symbols.arrow_up
                    return hdiff
            else:
                hdiff += " "
            i += 1
        while len(hdiff) < lim:
            hdiff += " "
        return hdiff

    @staticmethod
    def render_white(text: Text, color: str = "") -> Optional[Text]:
        out = Text().add(text).replace(' ', Token(symbols.whitespace.text, color)).replace('\n', Token(symbols.newline.text, color))
        return out

    def first_failure_diff(self, a_text: str, b_text: str | None, first_failure: int) -> List[Text]:
        if b_text is None:
            b_text = ""
        def get(vet, index):
            if index < len(vet):
                return DiffBuilder.render_white(vet[index])
            return ""

        a_render = a_text.splitlines(True)
        b_render = b_text.splitlines(True)
        first_a = get(a_render, first_failure)
        first_b = get(b_render, first_failure)
        out_a, out_b = DiffBuilder.colorize_2_lines_diff(Text().add(first_a), Text().add(first_b))
        if out_a is None:
            out_a = Text()
        if out_b is None:
            out_b = Text()
        greater = max(len(out_a), len(out_b))
        output: List[Text] = []
        width = self.width - 13
        output.append(Text().add(" ").add(out_a.ljust(greater)).trim_end(width).addf("g", " (esperado)"))
        output.append(Text().add(" ").add(out_b.ljust(greater)).trim_end(width).addf("r", " (recebido)"))
        diff = DiffBuilder.make_line_arrow_up(first_a, first_b)
        output.append(Text().add(" ").add(diff.ljust(greater)).trim_end(width).addf("b", " (primeiro)"))
        return output

    @staticmethod
    def colorize_2_lines_diff(la: Text | None, lb: Text | None, neut: str = "", exp: str = "g", rec: str = "r") -> Tuple[Text | None, Text | None]:
        pos = DiffBuilder.find_first_mismatch(la, lb)
        if la is not None:
            lat = la.get_text()
            la = Text().addf(neut, lat[0:pos]).addf(exp, lat[pos:])
        if lb is not None:
            lbt = lb.get_text()
            lb = Text().addf(neut, lbt[0:pos]).addf(rec, lbt[pos:])
        return la, lb

    @staticmethod
    def find_first_mismatch(line_a: Text | None, line_b: Text | None) -> int: 
        if line_a is None or line_b is None:
            return 0
        i = 0
        while i < len(line_a) and i < len(line_b):
            if line_a[i] != line_b[i]:
                return i
            i += 1
        return i

    # return a tuple of two strings with the diff and the index of the first mismatch line
    def render_diff(self, a_text: str, b_text: str | None) -> Tuple[List[Tuple[Text | None, Text | None]], int]:
        if b_text is None:
            b_text = ""
        a_lines = a_text.splitlines()
        b_lines = b_text.splitlines()
        output: list[tuple[Text | None, Text | None]] = []
        a_size = len(a_lines)
        b_size = len(b_lines)
        first_failure = -1
        max_size = max(a_size, b_size)

        # lambda function to return element in index i or empty if out of bounds
        def get(vet, index) -> Text | None:
            if index < len(vet):
                return Text(vet[index])
            return None
        
        expected_color = "g"
        received_color = "r" if a_text != "" else ""
        for i in range(max_size):
            a_data = get(a_lines, i)
            b_data = get(b_lines, i)
            if a_data is None or b_data is None or a_data != b_data:
                if first_failure == -1:
                    first_failure = i
                a_out, b_out = DiffBuilder.colorize_2_lines_diff(a_data, b_data, "y", expected_color, received_color)
                output.append((a_out, b_out))
            else:
                output.append((a_data, b_data))
        return output, first_failure



# the inclusion of the tests module is not meant to offer best practices for
# testing in general, but rather to support the `find_packages` example in
# setup.py that excludes installing the "tests" package

import unittest


from tko.game.task import TaskParser, Task


class TestSimple(unittest.TestCase):

    def test_reading_task(self):
        task: Task = TaskParser.parse_line("- [ ] [tarefa um](wiki/tarefa.md)", 0)
        assert task.key == "wiki/tarefa.md"
        assert task.link == "wiki/tarefa.md"
        assert task.self_grade == 0
        assert task.title == "tarefa um"

    def test_coding_local_task(self):
        task: Task = TaskParser.parse_line("- [ ] [minha @boneca quebrou](base/boneca/Readme.md)", 0)
        assert task.key == "boneca"
        assert task.link == "base/boneca/Readme.md"
        assert task.self_grade == 0
        assert task.title == "minha @boneca quebrou"


if __name__ == '__main__':
    unittest.main()



from typing import List, Optional
import os
import shutil
import subprocess

from ..run.diff_builder_down import DownDiff

from ..run.wdir import Wdir
from tko.util.consts import DiffCount, ExecutionResult
from tko.util.param import Param
from ..run.diff_builder_side import SideDiff
from tko.util.text import Text, Token

from tko.util.consts import Success
from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols

from tko.util.freerun import Free
from ..play.tester import Tester
from ..run.unit_runner import UnitRunner
from ..game.task import Task
from ..play.opener import Opener
from tko.settings.settings import Settings
from tko.util.consts import DiffMode
from tko.util.logger import Logger, LogAction, LoggerFS

class FilterMode:
    @staticmethod
    def deep_copy_and_change_dir():
        # path to ~/.tko_filter
        filter_path = os.path.join(os.path.expanduser("~"), ".tko_filter")

        # verify if filter command is available
        if shutil.which("filter_code") is None:
            print("ERROR: comando de filtragem não encontrado")
            print("Instale o feno com 'pip install feno'")
            exit(1)

        subprocess.run(["filter_code", "-rf", ".", "-o", filter_path])

        os.chdir(filter_path)

class Run:

    def __init__(self, settings: Settings, target_list: List[str], exec_cmd: Optional[str], param: Optional[Param.Basic]):
        self.settings = settings
        self.target_list: List[str] = target_list
        self.exec_cmd: Optional[str] = exec_cmd
        if param is None:
            self.param = Param.Basic()
        else:
            self.param = param
        self.wdir: Wdir = Wdir()
        self.wdir_builded = False
        self.__curses_mode: bool = False
        self.__lang = ""
        self.__task: Optional[Task] = None
        self.__opener: Optional[Opener] = None
        self.__autorun: bool = True

    def set_curses(self, value:bool=True, success: Success=Success.RANDOM):
        self.__curses_mode = value
        return self
   
    def set_lang(self, lang:str):
        self.__lang = lang
        return self
    
    def set_opener(self, opener: Opener):
        self.__opener = opener
        return self

    def set_autorun(self, value:bool):
        self.__autorun = value

    def set_task(self, task: Task):
        self.__task = task
        return self
    
    def get_task(self) -> Task:
        if self.__task is None:
            raise Warning("Task não definida")
        return self.__task

    def execute(self):
        if not self.wdir_builded:
            self.build_wdir()

        if self.__missing_target():
            return
        if self.__list_mode():
            return
        
        Logger.instance = Logger(logger_store=LoggerFS(self.settings))
        logger = Logger.instance
        rep_task = self.identity_rep_and_task_by_path()
        if rep_task is not None:
            rep, task_key = rep_task
            logger.set_rep(rep)
            if self.__task is None:
                task = Task()
                task.key = task_key
                self.__task = task

        if self.__free_run():
            return
        self.__show_diff()
        return

    def identity_rep_and_task_by_path(self) -> tuple[str, str] | None:
        rootdir = self.settings.app.get_rootdir()
        solver_path = os.path.abspath(self.wdir.get_solver().path_list[0])
        if not solver_path.startswith(rootdir):
            return None
        pieces = solver_path[len(rootdir):].split(os.sep)
        if len(pieces) < 2:
            return None
        rep = pieces[1]
        task = pieces[2] if len(pieces) > 2 else ""
        return (rep, task)
    
    def __remove_duplicates(self):
        # remove duplicates in target list keeping the order
        self.target_list = list(dict.fromkeys(self.target_list))

    def __change_targets_to_filter_mode(self):
        if self.param.filter:
            old_dir = os.getcwd()

            print(Text(" Entrando no modo de filtragem ").center(RawTerminal.get_terminal_size(), "═"))
            FilterMode.deep_copy_and_change_dir()  
            # search for target outside . dir and redirect target
            new_target_list = []
            for target in self.target_list:
                if ".." in target:
                    new_target_list.append(os.path.normpath(os.path.join(old_dir, target)))
                elif os.path.exists(target):
                    new_target_list.append(target)
            self.target_list = new_target_list

    def __print_top_line(self):
        if self.wdir is None:
            return

        print(Text().add(symbols.opening).add(self.wdir.resume()), end="")
        print(" [", end="")
        first = True
        for unit in self.wdir.get_unit_list():
            if first:
                first = False
            else:
                print(" ", end="")
            solver = self.wdir.get_solver()
            if solver is None:
                raise Warning("Solver vazio")
            unit.result = UnitRunner.run_unit(solver, unit)
            print(Text() + ExecutionResult.get_symbol(unit.result), end="")
        print("]")

    def __print_diff(self):
        if self.wdir is None or not self.wdir.has_solver():
            return
        
        if self.param.diff_count == DiffCount.QUIET:
            return
        
        if self.wdir.get_solver().compile_error:
            print(self.wdir.get_solver().error_msg)
            return
        
        results = [unit.result for unit in self.wdir.get_unit_list()]
        if ExecutionResult.EXECUTION_ERROR not in results and ExecutionResult.WRONG_OUTPUT not in results:
            return
        
        if not self.param.compact:
            for elem in self.wdir.unit_list_resume():
                print(elem)

        
        if self.param.diff_count == DiffCount.FIRST:
            # printing only the first wrong case
            wrong = [unit for unit in self.wdir.get_unit_list() if unit.result != ExecutionResult.SUCCESS][0]
            if self.param.diff_mode == DiffMode.DOWN:
                ud_diff_builder = DownDiff(RawTerminal.get_terminal_size(), wrong).to_insert_header()
                for line in ud_diff_builder.build_diff():
                    print(line)
            else:
                ss_diff_builder = SideDiff(RawTerminal.get_terminal_size(), wrong).to_insert_header(True)
                for line in ss_diff_builder.build_diff():
                    print(line)
            return

        if self.param.diff_count == DiffCount.ALL:
            for unit in self.wdir.get_unit_list():
                if unit.result != ExecutionResult.SUCCESS:
                    if self.param.diff_mode:
                        for line in ud_diff_builder.build_diff(unit):
                            print(line)
                    else:
                        for line in ss_diff_builder.build_diff(unit):
                            print(line)

    def build_wdir(self):
        self.wdir_builded = True
        self.__remove_duplicates()
        self.__change_targets_to_filter_mode()
        try:
            self.wdir = Wdir().set_curses(self.__curses_mode).set_lang(self.__lang).set_target_list(self.target_list).set_cmd(self.exec_cmd).build().filter(self.param)
        except FileNotFoundError as e:
            if self.wdir.has_solver():
                self.wdir.get_solver().error_msg += str(e)
                self.wdir.get_solver().compile_error = True
        return self.wdir

    def __missing_target(self) -> bool:
        if self.wdir is None:
            return False
        if not self.wdir.has_solver() and not self.wdir.has_tests():
            print(Text().addf("", "fail: ") + "Nenhum arquivo de código ou de teste encontrado.")
            return True
        return False
    
    def __list_mode(self) -> bool:
        if self.wdir is None:
            return False

        # list mode
        if not self.wdir.has_solver() and self.wdir.has_tests():
            print(Text(" Nenhum arquivo de código encontrado. Listando casos de teste.").center(RawTerminal.get_terminal_size(), Token("╌")), flush=True)
            print(self.wdir.resume())
            for line in self.wdir.unit_list_resume():
                print(line)
            return True
        return False

    def __free_run(self) -> bool:
        if self.wdir is None:
            return False
        if self.wdir.has_solver() and (not self.wdir.has_tests()) and not self.__curses_mode:
            if self.__task is not None:
                Logger.get_instance().record_event(LogAction.FREE, self.get_task().key)
            Free.free_run(self.wdir.get_solver(), show_compilling=False, to_clear=False, wait_input=False)
            return True
        return False

    def __create_opener_for_wdir(self) -> Opener:
        opener = Opener(self.settings)
        folders = []
        targets = ["."]
        if self.target_list:
            targets = self.target_list
        for f in targets:
            if os.path.isdir(f) and f not in folders:
                folders.append(f)
            else:
                folder = os.path.dirname(os.path.abspath(f))
                if folder not in folders:
                    folders.append(folder)
        opener.set_target(folders)
        solver_zero = self.wdir.get_solver().path_list[0]
        lang = solver_zero.split(".")[-1]
        opener.set_language(lang)
        return opener

    def __show_diff(self):
        if self.wdir is None:
            return
        
        if self.__curses_mode:
            cdiff = Tester(self.settings, self.wdir)
            if self.__task is not None:
                cdiff.set_task(self.__task)
            if self.__opener is not None:
                cdiff.set_opener(self.__opener)
            else:
                cdiff.set_opener(self.__create_opener_for_wdir())
            cdiff.set_autorun(self.__autorun)
            cdiff.run()
        else:
            print(Text(" Testando o código com os casos de teste ").center(RawTerminal.get_terminal_size(), "═"))
            self.__print_top_line()
            self.__print_diff()
            correct = [unit for unit in self.wdir.get_unit_list() if unit.result == ExecutionResult.SUCCESS]
            percent = (len(correct) * 100) // len(self.wdir.get_unit_list())
            if self.__task is not None:
                Logger.get_instance().record_event(LogAction.TEST, self.get_task().key, str(percent) + "%")



import unittest
import os

from .compare import compare_list

class Test:
    folder = "tests/03_run/cmd"

    def test_run_cmd(self, capsys):
        cmd = "g++ solver.cpp -o solver.out && ./solver.out"
        cmd2 = ["-w", "50", "-m", "test", "--cmd", cmd, "cases.tio"]
        compare_list(capsys, Test.folder, "out1", cmd2)


    def test_run_makefile(self, capsys):
        cmd = "make -s solver.out && ./solver.out"
        cmd2 = ["-w", "50", "-m", "test", "--cmd", cmd, "cases.tio"]
        compare_list(capsys, Test.folder, "out2", cmd2)



if __name__ == '__main__':
    unittest.main()



import unittest

from .compare import compare_text

class Test:
    folder = "tests/04_diff"
                
    def test_run_mixed_1(self, capsys):
        compare_text(capsys, Test.folder, "out1", "-w 80 -m test cases.tio solver.py -s")


    def test_run_mixed_2(self, capsys):
        compare_text(capsys, Test.folder, "out2", "-w 80 -m test cases.tio solver.py -d")

if __name__ == '__main__':
    unittest.main()



class GuiActions:
    tab = "↔TAB"
    leave   = "Sair"
    edit = "Editar"
    interact  = "Interagir"
    principal = "Principal"
    limite = "Limite"
    github = "Github"
    help = "Ajuda"
    download = "Baixar"
    activate = "Ativar"
    navigate = "←↓→"
    progress = "Quanto"
    uncheck = "Desmarcar"
    search = "Pesquisar"
    self_grade = " Como "
    palette = "Paleta"
    all = "Todos"

class GuiKeys:
    left = "a"
    right = "d"
    down = "s"
    up = "w"

    palette = "p"
    images = "I"
    down_task = "b"
    activate = "\n"
    key_help = "?"
    inc_prog = ">"
    prog_plus2 = "."
    dec_prog = "<"
    prog_less2 = ","
    inc_self = "+"
    inc_grade2 = "="
    dec_self = "-"
    dec_grade2 = "_"
    set_root_dir = "D"
    set_lang = "L"
    github_open = "g"
    key_quit = "q"
    edit= "e"
    colors = "C"
    borders = "B"
    search = "/"
    graph = "G"
    diff = "m"
    interact = "i"
    test = "\n"
    main = "\t"
    quit   = "q"
    lock = "t"
    limite  = "l"
    expand = "]"
    collapse = "["


from tko.play.floating import FloatingInputData, FloatingInput
from tko.util.text import Text, Token
from tko.util.symbols import symbols
from tko.play.keys import GuiKeys
from tko.play.flags import Flags
from tko.play.play_actions import PlayActions
from tko.play.floating_manager import FloatingManager

class PlayPalette:
    def __init__(self, actions: PlayActions):
        self.actions = actions
        self.app = self.actions.settings.app
        self.fman = self.actions.fman
        self.gui = self.actions.gui

    def command_pallete(self):
        options: list[FloatingInputData] = []

        def icon(value: bool):
            return Token("✓", "g") if value else Token("✗", "r")
        
        options.append(
            FloatingInputData(
                lambda: Text(" {} Tarefa: {y} para repositório local", symbols.action, "Baixar"),
                self.actions.down_task,
                GuiKeys.down_task
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Tarefa: Abrir {y} com a descrição", symbols.action, "GitHub"),
                self.actions.open_link,
                GuiKeys.github_open
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Tarefa: {y} arquivos na IDE", symbols.action, "Editar"),
                self.actions.open_code,
                GuiKeys.edit
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y}", symbols.action, "Ajuda"),
                self.gui.show_help,
                GuiKeys.key_help
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar barra de {y}", icon(Flags.flags.is_true()), "Flags"),
                Flags.flags.toggle,
                Flags.flags.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y}", icon(self.app.has_borders()), "Bordas"),
                self.app.toggle_borders,
                GuiKeys.borders
            )
        )
        
        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y}", icon(self.app.has_images()), "Imagens"),
                self.app.toggle_images, 
                GuiKeys.images
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y}", icon(Flags.percent.is_true()), "Percentual"),
                Flags.percent.toggle, 
                Flags.percent.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y} para completar a missão", icon(Flags.minimum.is_true()), "Mínimo"),
                Flags.minimum.toggle,
                Flags.minimum.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y} das tarefas", icon(Flags.reward.is_true()), "Recompensa"),
                Flags.reward.toggle, 
                Flags.reward.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mostrar {y}", icon(Flags.skills.is_true()), "Skills"),
                Flags.skills.toggle, 
                Flags.skills.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Modo {y}: Habilitar todas as tarefas", icon(Flags.admin.is_true()), "Admin"),
                Flags.admin.toggle,
                Flags.admin.get_keycode()
            )
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Gerar {y} de dependências", symbols.action, "Grafo"),
                self.actions.generate_graph,
                GuiKeys.graph
            ).set_exit_on_action(True)
        )

        options.append(
            FloatingInputData(
                lambda: Text(" {} Mudar {y} de download de rascunhos", symbols.action, "Linguagem"),
                self.gui.language.set_language,
                GuiKeys.set_lang
            ).set_exit_on_action(True)
        )

        self.fman.add_input(
            FloatingInput("^").set_text_ljust()
                      .set_header(" Selecione uma ação da lista ")
                      .set_options(options)
                      .set_exit_on_enter(False)
                      .set_footer(" Use Enter para aplicar e Esc para Sair ")
        )


class Drafts:
    ts_draft = r"""
function input(): string { let X: any = input; X.L = X.L || require("fs").readFileSync(0).toString().split(/\r?\n/); return X.L.shift(); } // _TEST_ONLY_
// function input(): string { let X: any = input; X.P = X.P || require("readline-sync"); return X.P.question() } // _FREE_ONLY_
function write(text: any, endl="\n") { process.stdout.write("" + text + endl); }
export {};

function main() {
    write("qxcode");
}
main();
"""[1:]

    js_draft = r"""
function input() { let X = input; X.L = X.L || require("fs").readFileSync(0).toString().split(/\r?\n/); return X.L.shift(); }  // _TEST_ONLY_
// function input() { let X = input; X.P = X.P || require("readline-sync"); return X.P.question() }  // _FREE_ONLY_
function write(text, endl="\n") { process.stdout.write("" + text + endl); }

function main() {
    write("qxcode");
}
main();
"""[1:]

    c_draft = r"""
#include <stdio.h>

int main() {
    puts("qxcode");
    return 0;
}

"""[1:]

    cpp_draft = r"""
#include <iostream>

int main() {
    std::cout << "qxcode\n";
}

"""[1:]

    java_draft = r"""
public class draft {
    public static void main(String args[]) {
        System.out.println("qxcode");
    }
}

"""[1:]

    go_draft = (
        r"package main""\n"
        r'import "fmt"'"\n"
        r"func main() {""\n"
        r'    fmt.Println("qxcode")''\n'
        r"}""\n"
    )


    drafts = {'c': c_draft, 'cpp': cpp_draft, 'ts': ts_draft, 'js': js_draft, 'java': java_draft, 'go': go_draft}


from tko.settings.settings import Settings
from tko.game.game import Game
from tko.settings.rep_settings import RepSource
from tko.play.play import Play
from tko.util.logger import Logger, LogAction
from typing import Dict
import os

class CmdPlay:
    @staticmethod
    def execute(rep_alias: str, settings: Settings):
        while True:
            rep_alias = settings.check_rep_alias(rep_alias)
            rep_source: RepSource = settings.get_rep_source(rep_alias)
            rep = settings.get_rep_data(rep_alias)
            rep_dir = rep.get_rep_dir()
            file = rep_source.get_file_or_cache(rep_dir)
            game = Game()
            game.parse_file(file)

            # passing a lambda function to the play class to save the settings
            play = Play(settings=settings, game=game, rep=rep)
            logger: Logger = Logger.get_instance()
            logger.set_rep(rep_alias)
            logger.record_event(LogAction.OPEN)
            print(f"Abrindo repositório de {rep_alias}")
            reload = play.play()
            if not reload:
                break
        logger.record_event(LogAction.QUIT)



from __future__ import annotations
from typing import List,  Any, Tuple, Union

class AnsiColor:
    enabled = True
    terminal_styles = {
        '.': '\033[0m', # Reset
        '*': '\033[1m', # Bold
        '/': '\033[3m', # Italic
        '_': '\033[4m', # Underline
        
        'k': '\033[30m', # Black
        'r': '\033[31m', # Red
        'g': '\033[32m', # Green
        'y': '\033[33m', # Yellow
        'b': '\033[34m', # Blue
        'm': '\033[35m', # Magenta
        'c': '\033[36m', # Cyan
        'w': '\033[37m', # White


        'K': '\033[40m', # Background black
        'W': '\033[47m', # Background white
        "R": '\033[41m', # Background red
        "G": '\033[42m', # Background green
        "Y": '\033[43m', # Background yellow
        "B": '\033[44m', # Background blue
        "M": '\033[45m', # Background magenta
        "C": '\033[46m', # Background cyan
    }

    @staticmethod
    def colour(modifiers: str, text: str) -> str:
        if not AnsiColor.enabled:
            return text
        output = ''
        for m in modifiers:
            val = AnsiColor.terminal_styles.get(m, '')
            if val != '':
                output += val
        output += text
        if len(modifiers) > 0:
            output += AnsiColor.terminal_styles.get('.', "")
        return output

class Token:
    def __init__(self, text: str = "", fmt: str = "", ):
        if not isinstance(text, str):
            raise TypeError("text must be a string")
        self.text = text
        self.fmt = fmt

    def __eq__(self, other: Any):
        if not isinstance(other, Token):
            return False
        return self.text == other.text and self.fmt == other.fmt

    def __len__(self):
        return len(self.text)
    
    def __add__(self, other: Any):
        return Text().add(self).add(other)

    def __str__(self):
        return f"({self.fmt}:{self.text})"

def RToken(fmt: str, text: str) -> Token:
    return Token(text, fmt)

class Text:
    def __init__(self, value: str = "", *args):
        if not isinstance(value, str):
            raise TypeError("value must be a string")
        self.data: List[Token] = []
        if value != "": 
            self.__process_placeholders(value, *args)
    
    def set_background(self, fmt: str):
        for d in self.data:
            lower_only = "".join([c for c in d.fmt if c.islower()])
            d.fmt = fmt + lower_only
        return self

    def clone(self):
        other = Text()
        other.data = [v for v in self.data]
        return other

    def setup(self, data: List[Token]):
        self.data = []
        for d in data:
            for c in d.text:
                self.data.append(Token(c, d.fmt))
        return self

    def split(self, sep: str) -> List[Text]:
        output = []
        current = Text()
        for d in self.data:
            if d.text == sep:
                output.append(current)
                current = Text()
            else:
                current.add(d)
        output.append(current)
        return output

    def slice(self, start: int | None = None, end: int | None = None) -> Text:
        if start is None:
            start = 0
        while start < 0:
            start = len(self) + start
        if end is None:
            end = len(self)
        while end < 0:
            end = len(self) + end
        output = Text()
        for i in range(start, end):
            output.add(self[i])
        return output

    def __getitem__(self, index: int) -> Token:
        if index < 0 or index >= len(self):
            raise IndexError("index out of range")
        return self.data[index]

    def __len__(self):
        return self.len()
    
    def __add__(self, other: Any):
        return Text().add(self).add(other)

    def __eq__(self, other: Any):
        if len(self.data) != len(other.data):
            return False
        for i in range(len(self.data)):
            if self.data[i] != other.data[i]:
                return False
        return True


    def __str__(self):
        output = ""
        for elem in self.resume():
            output += AnsiColor.colour(elem.fmt, elem.text)
        return output
    
    def resume(self) -> List[Token]:
        if len(self.data) == 0:
            return []
        
        new_data: List[Token] = [Token("", self.data[0].fmt)]
        for d in self.data:
            if d.fmt == new_data[-1].fmt:
                new_data[-1].text += d.text
            else:
                new_data.append(Token())
                new_data[-1].text = d.text
                new_data[-1].fmt = d.fmt
        return new_data

    # search for a value inside the tokens and replace it with a new value and fmt
    def replace(self, old: str, token: Token):
        index = 0
        while True:
            if index + len(old) > len(self.data):
                break
            sub = "".join([t.text for t in self.data[index:index + len(old)]])
            if sub == old:
                pre = self.data[:index]
                mid = [Token(t, token.fmt) for t in token.text]
                pos = self.data[index + len(old):]
                self.data = pre + mid + pos
                index += len(token.text)
            else:
                index += 1
        return self

    def plus(self, qtd: int) -> Text:
        output = Text()
        for i in range(qtd):
            output.add(self)
        return output

    def add(self, value: str | Token | Text | None):
        if value is None:
            return self
        if isinstance(value, str):
            if value != "":
                for c in value:
                    self.data.append(Token(c))
        elif isinstance(value, Token):
            if value.text != "":
                for c in value.text:
                    self.data.append(Token(c, value.fmt))
        elif isinstance(value, Text):
            self.data += value.data
        else:
            raise TypeError("unsupported type '{}'".format(type(value)))
        return self
    
    def addf(self, fmt: str, value: str | Token | Text | None):
        if isinstance(value, str):
            self.add(Token(value, fmt))
        elif isinstance(value, Token):
            self.add(Token(value.text, fmt))
        elif isinstance(value, Text):
            self.add(Token(value.get_text(), fmt))
        return self

    def ljust(self, width: int, filler: Token = Token(" ")):
        total = self.len()
        char = " " if filler.text == "" else filler.text[0]
        fmt = filler.fmt
        if total < width:
            suffix = [Token(char, fmt) for _ in range(width - total)]
            self.data = self.data + suffix
        return self
    
    def rjust(self, width: int, filler: Token = Token(" ")):
        total = self.len()
        char = " " if filler.text == "" else filler.text[0]
        fmt = filler.fmt
        if total < width:
            prefix = [Token(char, fmt) for _ in range(width - total)]
            self.data = prefix + self.data
        return self
    
    def center(self, width: int, filler: Token | str  = " "):
        if isinstance(filler, str):
            filler = Token(filler)
        total = self.len()
        char = " " if filler.text == "" else filler.text[0]
        fmt = filler.fmt
        if total < width:
            left = (width - total) // 2
            right = width - total - left
            prefix = [Token(char, fmt) for _ in range(left)]
            suffix = [Token(char, fmt) for _ in range(right)]
            self.data = prefix + self.data + suffix
        return self

    def fold_in(
        self, width: int, sep: str | Token = " ", left_border: str | Token = " ", right_border: str | Token  = " "
    ) -> Text:
        if isinstance(sep, str):
            sep = Token(sep)
        if isinstance(left_border, str):
            left_border = Token(left_border)
        if isinstance(right_border, str):
            right_border = Token(right_border)

        available = width - len(left_border) - len(right_border)
        if self.len() > available:
            self.trim_end(available)
            return Text() + left_border + self + right_border
        missing = width - self.len() - 2
        left = missing // 2
        right = missing - left
        filler_left = Text().add(sep).plus(left)
        filler_right = Text().add(sep).plus(right)
        return Text() + left_border + filler_left + self + filler_right + right_border
    
    def len(self):
        return len(self.data)
    
    def get_data(self):
        return self.data
    
    def get_text(self) -> str:
        return "".join([t.text for t in self.data])

    def cut_begin(self, qtd: int):
        if qtd > len(self.data):
            self.data = []
        else:
            self.data = self.data[qtd:]
        return self

    def trim_spaces(self, limit: int):
        return self

    def trim_end(self, width: int):
        if self.len() > width:
            self.data = self.data[:width]
        return self

    def join(self, array: List[Text]):
        out = Text()
        for i, a in enumerate(array):
            if i != 0:
                out.add(self)
            out.add(a)
        return out

    @staticmethod
    def __preprocess(data):
        # Primeira etapa: substitui << por \a
        result1 = []
        i = 0
        while i < len(data):
            if i < len(data) - 1 and data[i] == '{' and data[i + 1] == '{':
                result1.append('\a')  # Usa o marcador \a para << 
                i += 2  # Pula o próximo '<'
            else:
                result1.append(data[i])
                i += 1

        # Segunda etapa: substitui >> por \b, de trás para frente
        result2 = []
        i = len(result1) - 1
        while i >= 0:
            if i > 0 and result1[i] == '}' and result1[i - 1] == '}':
                result2.append('\b')  # Usa o marcador \b para >>
                i -= 2  # Pula o próximo '>'
            else:
                result2.append(result1[i])
                i -= 1

        # Inverter result2 para obter o resultado final
        final_result = ''.join(result2[::-1])
        return final_result

    @staticmethod
    def __extract(data):
        data = Text.__preprocess(data)  # Preprocessar a string
        texts: list[str] = [""]
        placeholders: list[str] = []
        destiny = texts  # Começa preenchendo textos"

        for c in data:
            if c == '{':
                placeholders.append("")  # Inicia um novo placeholder
                destiny = placeholders  # Muda o destino para placeholders
            elif c == '}':
                texts.append("")  # Volta para o texto
                destiny = texts
            else:
                # Substitui \a e \b por < e >
                if c == '\a':
                    c = '{'
                elif c == '\b':
                    c = '}'
                # Adiciona o caractere ao destino atual
                destiny[-1] += c

        # Garantir que o tamanho dos textos e placeholders sejam iguais
        while len(texts) > len(placeholders):
            placeholders.append("")

        return texts, placeholders

    def __process_placeholders(self, text, *args) -> None:
        params = list(args)
        texts, placeholders = Text.__extract(text)
        while len(params) < len(placeholders):
            params.append("")
        # Aplicar a função fn aos placeholders e montar o resultado final
        for i in range(len(texts)):
            value = texts[i]
            self.add(value)
            
            fmt = placeholders[i]
            value = params[i]
            if fmt == "":
                self.add(value)
            else:
                self.addf(fmt, value)

if __name__ == "__main__":
    # print(Text("comida {c:gelada}"))
    # print(Text("comida {c:gelada} e {r:quente}"))
    print(Text("a {b} está com {}", "agua", Text("{g}", "capim")))
    



from tko.game.task import Task
from tko.game.quest import Quest
from tko.play.tasktree import TaskTree
from tko.play.flags import Flags
from tko.play.floating import Floating
from tko.play.floating_manager import FloatingManager
from tko.play.input_manager import InputManager
from tko.play.fmt import Fmt
import curses
from typing import List

class Search:
    def __init__(self, tree: TaskTree, fman: FloatingManager):
        self.tree = tree
        self.game = tree.game
        self.fman = fman
        self.search_mode: bool = False
        self.backup_expanded: List[str] = []
        self.backup_index_selected: str = ""
        self.backup_admin_mode: bool = False

    def toggle_search(self):
        self.search_mode = not self.search_mode
        if self.search_mode:
            self.backup_expanded = [v for v in self.tree.expanded]
            self.backup_index_selected = self.tree.selected_item
            self.backup_admin_mode = Flags.admin.is_true()
            self.tree.update_tree(admin_mode=True)
            self.tree.process_expand()
            self.tree.process_expand()
            self.fman.add_input(Floating(">v").warning().put_text("Digite o texto\nNavegue até o elemento desejado\ne aperte Enter"))
    
    def finish_search(self):
        self.search_mode = False
        unit = self.tree.get_selected()
        self.tree.selected_item = ""
        self.tree.search_text = ""
        if self.backup_admin_mode == False:
            self.tree.update_tree(admin_mode=False)
        self.tree.reload_sentences()
    
        found = False
        for i, item in enumerate(self.tree.items):
            if item == unit:
                self.tree.selected_item = item.get_key()
                found = True
                break

        if not found:
            self.fman.add_input(Floating(">v").warning().put_text("Elemento não acessível no modo normal.\nEntrando no modo Admin\npara habilitar acesso"))
            Flags.admin.toggle()
            self.tree.update_tree(True)
            self.tree.reload_sentences()
        
        self.tree.process_collapse()
        self.tree.process_collapse()
        self.tree.selected_item = unit.get_key()
        if isinstance(unit, Task):
            for cluster_key in self.game.available_clusters:
                cluster = self.game.clusters[cluster_key]
                for quest in cluster.quests:
                    for task in quest.get_tasks():
                        if task == unit:
                            self.tree.expanded = [cluster.key, quest.key]
        elif isinstance(unit, Quest):
            for cluster_key in self.game.available_clusters:
                cluster = self.game.clusters[cluster_key]
                for quest in cluster.quests:
                    if quest == unit:
                        self.tree.expanded = [cluster.key]
        self.tree.reload_sentences()
        for i, item in enumerate(self.tree.items):
            if item== unit:
                self.tree.selected_item = item.get_key()
                break

    # update index to match the first item that matches the search
    def update_index(self):
        filtered, first = self.tree.filter_by_search()
        self.tree.selected_item = first if first is not None else self.tree.selected_item
        # if first is not None:
        #     for i, item in enumerate(self.tree.items):
        #         if item.get_key() == first:
        #             self.tree.selected_item = item.get_key()
        #             break


    def process_search(self, key):
        if key == 27:
            self.search_mode = False
            self.tree.search_text = ""
            self.tree.expanded = [v for v in self.backup_expanded]
            self.tree.selected_item = self.backup_index_selected
        elif key == ord("\n"):
            self.finish_search()
    
        elif key == curses.KEY_UP:
            self.tree.move_up()
        elif key == curses.KEY_DOWN:
            self.tree.move_down()
        elif key == InputManager.backspace1 or key == InputManager.backspace2 or key == InputManager.delete:
            self.tree.search_text = self.tree.search_text[:-1]
            self.update_index()
        elif key >= 32 and key < 127:
            self.tree.search_text += chr(key).lower()
            self.update_index()



from typing import List, Any, Dict, Tuple, Union, Set

from tko.settings.app_settings import AppSettings
from tko.settings.rep_settings import RepData

from tko.util.text import Text, Token
from tko.util.to_asc import SearchAsc, uni_to_asc
from tko.play.flags import Flags
from tko.game.game import Game
from tko.game.cluster import Cluster
from tko.game.quest import Quest
from tko.game.task import Task
from tko.play.border import Border
from tko.play.colors import Colors
from tko.util.symbols import symbols
from tko.settings.settings import Settings
from tko.play.floating_manager import FloatingManager
from tko.play.floating import Floating
from tko.play.grade_message import GradeMessage
from tko.game.tree_item import TreeItem

import os

class TaskTree:

    def __init__(self, settings: Settings, game: Game, rep: RepData, fman: FloatingManager):
        self.settings = settings
        self.app = settings.app
        self.game = game
        self.rep = rep
        self.fman = fman
        self.style = Border(settings.app)
        self.colors = settings.colors
        self.items: List[TreeItem] = []
        self.selected_item: str = ""
        self.index_begin = 0
        self.max_title = 0
        self.search_text = ""
        self.load_from_rep()
        self.update_tree(admin_mode=Flags.admin.is_true(), first_loop=True)
        self.reload_sentences()

    def load_from_rep(self):
        self.new_items: List[str] = [v for v in self.rep.get_new_items()]
        self.expanded: List[str] = [v for v in self.rep.get_expanded()]
        self.selected_item = self.rep.get_selected()

        tasks = self.rep.get_tasks()
        for key, serial in tasks.items():
            if key in self.game.tasks:
                self.game.tasks[key].load_from_db(serial)

    def save_on_rep(self):
        self.rep.set_expanded(self.expanded)
        self.rep.set_new_items([x for x in set(self.new_items)])
        self.rep.set_selected(self.selected_item)
        tasks = {}
        for t in self.game.tasks.values():
            if not t.is_db_empty():
                tasks[t.key] = t.save_to_db()
        self.rep.set_tasks(tasks)


    def update_tree(self, admin_mode: bool, first_loop: bool = False):
        if not admin_mode:
            old_reachable = self.game.available_clusters + self.game.available_quests

        self.game.update_reachable_and_available(admin_mode)
            
        if not admin_mode:
            available_keys = self.game.available_quests + self.game.available_clusters
            for key in self.expanded:
                if key not in available_keys:
                    self.expanded.remove(key)
            if not first_loop:
                for key in available_keys:
                    if key not in old_reachable and key not in self.new_items:
                        self.new_items.append(key)

        # remove expanded items from new
        self.new_items = [item for item in self.new_items if item not in self.expanded]

    def update_max_title(self):
        min_value = 50
        items = []
        for c in self.game.clusters.values():
            if c.key in self.game.available_clusters:
                items.append(len(c.title))
                if c.key in self.expanded:
                    for q in [q for q in c.quests if q.key in self.game.available_quests]:
                        items.append(len(q.title) + 2)
                        if q.key in self.expanded:
                            for t in q.get_tasks():
                                items.append(len(t.title) + 6)
        self.max_title = max(items)
        if self.max_title < min_value:
            self.max_title = min_value

    def str_task(self, focus_color: str, t: Task, lig_cluster: str, lig_quest: str, quest_reachable: bool, min_value=1) -> Text:
        # downloadable_in_focus = False
        rootdir = self.app._rootdir
        down_symbol = Token(" ")
        in_focus = focus_color != ""
        down_symbol = symbols.cant_download
        rep_dir = os.path.join(self.app.get_rootdir(), self.rep.alias)
        if t.is_downloadable() and rootdir != "":
            if t.is_downloaded_for_lang(rep_dir, self.rep.get_lang()):
                down_symbol = symbols.downloaded
                # if in_focus:
                #     downloadable_in_focus = True
            else:
                down_symbol = symbols.to_download

        color_aval = "" if quest_reachable else "r"

        output = Text()
        output.add(" ").addf(color_aval, lig_cluster)
        output.add(" ")
        output.addf(color_aval, lig_quest)
        output.add(down_symbol).add(" ")
        output.add(t.get_prog_symbol()).add(" ")
        output.add(GradeMessage().grade_to_emojis(t.self_grade))

        if in_focus:
            output.add(self.style.roundL(focus_color))
        else:
            output.add(" ")

        color = ""
        if in_focus:
            color = "k" + focus_color

        done = color + "g"
        todo = color
        perc = t.progress
        output.add(self.style.build_bar(t.title, perc / 100, len(t.title), done, todo, round=False))
        # output.addf(color, t.title)

        if in_focus:
            output.add(self.style.roundR(focus_color))
        else:
            output.add(" ")


        if Flags.reward.is_true():
            xp = ""
            for s, v in t.skills.items():
                xp += f" +{s}:{v}"
            output.addf(self.colors.task_skills, xp)
            
        return output

    def str_quest(self, has_kids: bool, focus_color: str, q: Quest, lig: str) -> Text:
        con = "━─"
        if q.key in self.expanded and has_kids:
            con = "─┯"

        color_reachable = "" if q.is_reachable() else "r"
        output: Text = Text().addf(color_reachable, " " + lig + con)

        in_focus = focus_color != ""
        if in_focus:
            output.add(self.style.roundL(focus_color))
        else:
            output.add(" ")

        color = ""
        if in_focus:
            color = "k" + focus_color

        title = q.title
        title = title.ljust(self.max_title - 2, ".")

        done = color + self.colors.task_text_done
        todo = color + self.colors.task_text_todo
        output.add(self.style.build_bar(title, q.get_percent() / 100, len(title), done, todo, round=False))

        if in_focus:
            output.add(self.style.roundR(focus_color))
        else:
            output.add(" ")

        if Flags.percent.is_true():
            output.add(" ").add(q.get_resume_by_percent())
        else:
            output.add(" ").add(q.get_resume_by_tasks())

        if Flags.minimum.is_true():
            output.add(" ").add(q.get_requirement())

        if Flags.reward.is_true():
            xp = ""
            for s, v in q.skills.items():
                xp += f" +{s}:{v}"
            output.addf(self.colors.task_skills, " " + xp)

        if q.key in self.new_items:
            output.addf(self.colors.task_new, " [new]")

        return output


    def str_cluster(self, has_kids: bool, focus_color: str, cluster: Cluster) -> Text:
        output: Text = Text()
        opening = "━─"
        if cluster.key in self.expanded and has_kids:
            opening = "─┯"
        color_reachable = "" if cluster.is_reachable() else "r"
        output.addf(color_reachable, opening)

        color = ""
        if focus_color != "":
            color = "k" + focus_color
        title = cluster.title

        title = cluster.title.ljust(self.max_title, ".")
        if focus_color != "":
            output.add(self.style.roundL(focus_color))
        else:
            output.add(" ")

        done = color + self.colors.task_text_done
        todo = color + self.colors.task_text_todo

        output.add(self.style.build_bar(title, cluster.get_percent() / 100, len(title), done, todo, round=False))

        if focus_color != "":
            output.add(self.style.roundR(focus_color))
        else:
            output.add(" ")


        if Flags.percent.is_true():
            output.add(" ").add(cluster.get_resume_by_percent())
        else:
            output.add(" ").add(cluster.get_resume_by_quests())
        if cluster.key in self.new_items:
            output.addf(self.colors.task_new, " [new]")

        return output


    # def add_in_search(self, item: Any, sentence: Sentence) -> bool:
    #     if self.search_text == "":
    #         self.items.append(Entry(item, sentence))
    #         return True
        
    #     matcher = SearchAsc(self.search_text)
    #     pos = matcher.find(sentence.get_text())
    #     found = pos != -1
    #     if found:
    #         for i in range(pos, pos + len(self.search_text)):
    #             sentence.data[i].fmt = "Y"

    #     if isinstance(item, Task):
    #         if found:
    #             self.items.append(Entry(item, sentence))
    #             return True
    #     elif isinstance(item, Quest):
    #         if found:
    #             self.items.append(Entry(item, sentence))
    #             return True
    #         for t in item.get_tasks():
    #             if matcher.inside(t.title):
    #                 self.items.append(Entry(item, sentence))
    #                 return True
    #     elif isinstance(item, Cluster):
    #         cluster: Cluster = item
    #         if matcher.inside(cluster.title):
    #             self.items.append(Entry(cluster, sentence))
    #             return True
    #         for q in cluster.quests:
    #             if matcher.inside(q.title):
    #                 self.items.append(Entry(item, sentence))
    #                 return True
    #             for t in q.get_tasks():
    #                 if matcher.inside(t.title):
    #                     self.items.append(Entry(item, sentence))
    #                     return True
    #     return False

    def get_focus_color(self, item: Union[Quest, Cluster]) -> str:
        if not item.is_reachable() and Flags.admin.is_true():
            return "R"
        return self.colors.focused_item

    def filter_by_search(self) -> Tuple[Set[str], str | None]:
        matches: Set[str] = set()
        search = SearchAsc(self.search_text)
        first: None | str = None
        for cluster in self.game.clusters.values():
            if search.inside(cluster.title):
                matches.add(cluster.key)
                first = first or cluster.key
            for quest in cluster.quests:
                if search.inside(quest.title):
                    first = first or quest.key
                    matches.add(cluster.key)
                    matches.add(quest.key)
                for task in quest.get_tasks():
                    if search.inside(task.title):
                        first = first or task.key
                        matches.add(cluster.key)
                        matches.add(quest.key)
                        matches.add(task.key)
        return (matches, first)

    def try_add(self, filtered: set[str], matcher: SearchAsc, item: TreeItem):
        if self.search_text == "":
            self.items.append(item)
            return True
        if item.key in filtered:
            pos = matcher.find(item.sentence.get_text())
            found = pos != -1
            if found:
                for i in range(pos, pos + len(self.search_text)):
                    item.sentence.data[i].fmt = "Y"
            self.items.append(item)
            return True
        return False

    def reload_sentences(self):
        self.update_max_title()
        self.items = []
        available_quests = self.game.available_quests
        available_clusters = self.game.available_clusters

        filtered, _ = self.filter_by_search()
        matcher = SearchAsc(self.search_text)

        clusters = [self.game.clusters[key] for key in available_clusters if key in filtered]
        for cluster in clusters:
            quests = [q for q in cluster.quests if q.key in available_quests if q.key in filtered]
            focus_color = self.get_focus_color(cluster) if self.selected_item == cluster.get_key() else ""
            cluster.sentence = self.str_cluster(len(quests) > 0, focus_color, cluster)

            self.try_add(filtered, matcher, cluster)

            if cluster.key not in self.expanded:  # adicionou o cluster, mas não adicione as quests
                continue

            for q in quests:
                tasks =[t for t in q.get_tasks() if t.key in filtered]
                lig = "├" if q != quests[-1] else "╰"
                focus_color = self.get_focus_color(q) if self.selected_item == q.get_key() else ""
                q.sentence = self.str_quest(len(tasks) > 0, focus_color, q, lig)

                # self.items.append(Entry(q, sentence))
                self.try_add(filtered, matcher, q)
                if q.key in self.expanded:
                    for t in tasks:
                        ligc = "│" if q != quests[-1] else " "
                        ligq = "├ " if t != tasks[-1] else "╰ "
                        min_value = 7 if q.tmin is None else q.tmin
                        focus_color = self.get_focus_color(q) if self.selected_item == t.get_key() else ""
                        t.sentence = self.str_task(focus_color, t, ligc, ligq, q.is_reachable(), min_value)
                        self.try_add(filtered, matcher, t)


    def process_collapse(self):
        if any([q in self.expanded for q in self.game.available_quests]):
            self.expanded = [key for key in self.expanded if key not in self.game.available_quests]
        else:
            self.expanded = []

    def process_expand(self):
        # if any cluster outside expanded
        expand_clusters = False
        for ckey in self.game.available_clusters:
            if ckey not in self.expanded:
                expand_clusters = True
        if expand_clusters:
            for ckey in self.game.available_clusters:
                if ckey not in self.expanded:
                    self.expanded.append(ckey)
        else:
            for qkey in self.game.available_quests:
                if qkey not in self.expanded:
                    self.expanded.append(qkey)

    def get_selected_index(self) -> int:
        for i, item in enumerate(self.items):
            if item.key == self.selected_item:
                return i
        return 0

    def get_selected(self) -> TreeItem:
        index = self.get_selected_index()
        return self.items[index]

    def mass_mark(self):
        obj = self.get_selected()
        if isinstance(obj, Cluster):
            cluster: Cluster = obj
            if cluster.key not in self.expanded:
                self.expanded.append(cluster.key)
                return
            full_open = True
            for q in cluster.quests:
                if q.key in self.game.available_quests and q.key not in self.expanded:
                    self.expanded.append(q.key)
                    full_open = False
            if not full_open:
                return

            value = None
            for q in obj.quests:
                for t in q.get_tasks():
                    if value is not None:
                        t.set_grade(value)
                    else:
                        value = 10 if t.self_grade < 10 else 0
                        t.set_grade(value)
        elif isinstance(obj, Quest):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
            else:
                value = None
                for t in obj.get_tasks():
                    if value is not None:
                        t.set_grade(value)
                    else:
                        value = 10 if t.self_grade < 10 else 0
                        t.set_grade(value)
        elif isinstance(obj, Task):
            obj.set_grade(10 if obj.self_grade < 10 else 0)

    def set_grade(self, grade: int):

        obj = self.get_selected()
        if isinstance(obj, Task):
            obj.set_grade(grade)
            self.fman.add_input(
                Floating("v").warning().set_header(" Auto avaliação ").set_text_ljust().set_content(GradeMessage().format(grade).split("\n"))
            )

    def set_progress(self, prog: int):

        obj = self.get_selected()
        if isinstance(obj, Task):
            obj.progress = prog

    def inc_progress(self):
        obj = self.get_selected()
        if isinstance(obj, Task):
            progress = obj.progress + 10
            if progress > 100:
                progress = 100
            self.set_progress(progress)

    def dec_progress(self):
        obj = self.get_selected()
        if isinstance(obj, Task):
            progress = obj.progress - 10
            if progress < 0:
                progress = 0
            self.set_progress(progress)

    def inc_self_grade(self):
        obj = self.get_selected()
        if isinstance(obj, Task):
            grade = obj.self_grade + 1
            if grade >= 10:
                grade = 10
            self.set_grade(grade)
        else:
            self.unfold(obj)

    
    def dec_self_grade(self):
        obj = obj = self.get_selected()
        if isinstance(obj, Task):
            grade = obj.self_grade - 1
            if grade == -1:
                grade = 0
            self.set_grade(grade)
        else:
            self.fold(obj)

    def set_selected_by_index(self, index: int):
        if index < 0:
            index = 0
        if index >= len(self.items):
            index = len(self.items) - 1
        self.selected_item = self.items[index].key

    def arrow_right(self):
        index = self.get_selected_index()
        obj = self.items[index]
        if isinstance(obj, Cluster):
            if not self.unfold(obj):
                self.set_selected_by_index(index + 1)
        elif isinstance(obj, Quest):
            if not self.unfold(obj):
                while True:
                    index += 1
                    if index >= len(self.items):
                        break
                    obj = self.items[index]
                    if isinstance(obj, Cluster) or isinstance(obj, Quest):
                        break
                self.set_selected_by_index(index)
        elif isinstance(obj, Task):
            while True:
                obj = self.items[index]
                if isinstance(obj, Quest) or isinstance(obj, Cluster):
                    break
                if index == len(self.items) - 1:
                    break
                index += 1
            self.set_selected_by_index(index)

    def arrow_left(self):
        index = self.get_selected_index()
        obj = self.items[index]
        if isinstance(obj, Quest):
            if not self.fold(obj):
                while True:
                    if self.selected_item == 0:
                        break
                    index -= 1
                    obj = self.items[index]
                    if isinstance(obj, Cluster) or isinstance(obj, Quest) and obj.key in self.expanded:
                        break
                    if self.selected_item == 0:
                        break
        elif isinstance(obj, Cluster):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
                for q in obj.quests:
                    try:
                        self.expanded.remove(q.key)
                    except ValueError:
                        pass
            else:
                while True:
                    if self.selected_item == 0:
                        break
                    index -= 1
                    obj = self.items[index]
                    if isinstance(obj, Cluster) or isinstance(obj, Quest):
                        break
                    if self.selected_item == 0:
                        break
        elif isinstance(obj, Task):
            while True:
                obj = self.items[index]
                if isinstance(obj, Quest):
                    break
                index -= 1
        self.set_selected_by_index(index)

    def unfold(self, obj: TreeItem) -> bool:
        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            if obj.key not in self.expanded:
                self.expanded.append(obj.key)
                return True
        return False

    def fold(self, obj: TreeItem) -> bool:
        if isinstance(obj, Quest) or isinstance(obj, Cluster):
            if obj.key in self.expanded:
                self.expanded.remove(obj.key)
                return True
        return False

    def toggle(self, obj: Union[Quest, Cluster]):
            if not self.fold(obj):
                self.unfold(obj)

    def get_senteces(self, dy):
        index = self.get_selected_index()
        if len(self.items) < dy:
            self.index_begin = 0
        else:
            if index < self.index_begin:  # subiu na tela
                self.index_begin = index
            elif index >= dy + self.index_begin:  # desceu na tela
                self.index_begin = index - dy + 1

        sentences: List[Text] = []
        for i in range(self.index_begin, len(self.items)):
            sentences.append(self.items[i].sentence)
        return sentences

    def move_up(self):
        index = self.get_selected_index()
        self.set_selected_by_index(index - 1)

    def move_down(self):
        index = self.get_selected_index()
        self.set_selected_by_index(min(len(self.items) - 1, index + 1))



from typing import List
from .frame import Frame
from ..util.text import Text, Token
from .fmt import Fmt
import curses
from typing import Callable
from tko.play.input_manager import InputManager
from tko.util.symbols import symbols

class Floating:
    def __init__(self, _align=""):
        self._frame = Frame(0, 0)
        self._content: List[Text] = []
        self._type = "warning"
        self._enable = True
        self._extra_exit: List[int] = []
        self._exit_fn = None
        self._exit_key = None
        self._centralize = True
        self._floating_align = _align

    def disable(self):
        self._enable = False

    def set_text_ljust(self):
        self._centralize = False
        return self

    def set_text_center(self):
        self._centralize = True
        return self

    def set_frame_align(self, _align: str):
        self._floating_align = _align
        return self

    def set_header(self, text: str):
        self._frame.set_header(Text().addf("/", text), "")
        return self
    
    def set_header_sentence(self, sentence: Text):
        self._frame.set_header(sentence, "")
        return self
    
    def set_footer(self, text: str):
        self._frame.set_footer(Text().addf("/", text), "")
        return self
    
    def set_exit_key(self, key: str):
        self._exit_key = ord(key)
        return self

    def set_exit_fn(self, fn):
        self._exit_fn = fn
        return self

    def _set_xy(self, dy, dx):
        valid = "<>^v"
        for c in self._floating_align:
            if c not in valid:
                raise ValueError("Invalid align " + c)

        lines, cols = Fmt.get_size()

        x = (cols - dx) // 2
        if "<" in self._floating_align:
            x = 1
        elif ">" in self._floating_align:
            x = cols - dx - 3

        y = (lines - dy) // 2
        if "^" in self._floating_align:
            y = 1
        elif "v" in self._floating_align:
            y = lines - dy - 5

        self._frame.set_pos(y, x)
        return self
            
    def is_enable(self):
        return self._enable

    def calc_dy_dx(self):
        header_len = self._frame.get_header().len()
        footer_len = self._frame.get_footer().len()
        data = [x.len() for x in self._content] + [header_len, footer_len]
        max_dx = max(data)
        dx = max_dx
        dy = len(self._content)
        return dy, dx

    def setup_frame(self):
        dy, dx = self.calc_dy_dx()
        self._frame.set_inner(dy, dx)
        self._set_xy(dy, dx)
        self._frame.set_fill()

    def put_text(self, text: str):
        lines = text.split("\n")
        for line in lines:
            self._content.append(Text().add(line))
        return self

    def put_sentence(self, sentence: Text):
        for line in sentence.split("\n"):
            self._content.append(line)
        return self
    
    def set_content(self, content: List[str]):
        self._content = [Text().add(x) for x in content]
        return self

    def _set_default_footer(self):
        if self._frame.get_footer().len() == 0:
            label = Text().addf("/", " Pressione espaço ")
            self._frame.set_footer(label, "", "─", "─")
        return self

    def _set_default_header(self):
        if self._frame.get_header().len() == 0:
            if self._type == "warning":
                self.set_header(" Aviso ")
            elif self._type == "error":
                self.set_header(" Erro ")

    def warning(self):
        self._type = "warning"
        self._frame.set_border_color("y")
        return self
    
    def error(self):
        self._type = "error"
        self._frame.set_border_color("r")
        return self

    def draw(self):
        self._set_default_header()
        self._set_default_footer()
        self.setup_frame()
        self._frame.draw()
        self.write_content()

    def write_content(self):
        y = 0
        for line in self._content:
            x = 0
            if self._centralize:
                x = (self._frame.get_dx() - line.len()) // 2
            self._frame.write(y, x, line)
            y += 1
        return self

    def get_input(self) -> int:
        self.draw()
        key: int = Fmt.getch()
        key = InputManager.fix_cedilha(Fmt.get_screen(), key)
        if self._type == "warning" or self._type == "error":
            if key < 300:
                self._enable = False
                if self._exit_fn is not None:
                    self._exit_fn()
                if self._exit_key is not None:
                    return self._exit_key
                if key == ord(" ") or key == 27:
                    return -1
                return key
        
        return -1
        
class FloatingInputData:
    def __init__(self, label: Callable[[], Text], action: Callable[[], None], shortcut: str = ""):
        self.label = label
        self.action = action
        self.shortcut = shortcut
        self.exit_on_action = False


    def set_exit_on_action(self, value: bool):
        self.exit_on_action = value
        return self

class FloatingInput(Floating):
    def __init__(self, _align=""):
        super().__init__(_align)
        self._index = 0
        self._options: List[FloatingInputData] = []
        self._frame.set_border_color("m")
        self._exit_on_action = True
        self.right_dx = 5 # shortcut space
        self.search_text: List[str] = []

    def calc_dy_dx(self):
        dy, dx = super().calc_dy_dx()
        dy += len(self._options) + 2
        for option in self._options:
            dx = max(dx, len(option.label()))
        return dy, dx + self.right_dx
    
    def set_exit_on_enter(self, value: bool):
        self._exit_on_action = value
        return self

    def match_search(self, index: int):
        return "".join(self.search_text) in self._options[index].label().get_text().lower()

    def next_option(self):
        if not self.match_search(self._index):
            self._index = 0
        steps = len(self._options)
        index = self._index
        while steps > 0:
            index = (index + 1) % len(self._options)
            if self.match_search(index):
                self._index = index
                return
            steps -= 1
    
    def prev_option(self):
        if not self.match_search(self._index):
            self._index = 0
        steps = len(self._options)
        index = self._index
        while steps > 0:
            index = (index - 1) % len(self._options)
            if self.match_search(index):
                self._index = index
                return
            steps -= 1

    def write_content(self):
        options: List[Text] = []
        dx = self._frame.get_dx() - self.right_dx
        for i, option in enumerate(self._options):
            if not "".join(self.search_text) in option.label().get_text().lower():
                continue
            text = Text().add(option.label()).ljust(dx)
            if option.shortcut != "":
                if len(option.shortcut) > 1:
                    text.add(" " + option.shortcut)
                else:
                    text.add(f" [{option.shortcut}]")
            fmt = "M" if i == self._index else ""
            text.set_background(fmt)
            options.append(text)
        
        self._frame.write(0, 0, Text("Busca: ") + "".join(self.search_text) + symbols.cursor)
        y = 1
        for line in self._content + options:
            x = 0
            if self._centralize:
                x = (self._frame.get_dx() - line.len()) // 2
            self._frame.write(y, x, line)
            y += 1

        return self

    def set_options(self, options: List[FloatingInputData]):
        self._options = options
        return self
    
    def set_default_index(self, index: int):
        self._index = index
        return self

    def update_index(self):
        for i, _ in enumerate(self._options):
            if self.match_search(i):
                self._index = i
                return

    def get_input(self) -> int:
        self.draw()
        key: int = Fmt.getch()
        key = InputManager.fix_cedilha(Fmt.get_screen(), key)
        
        if key == curses.KEY_UP:
            self.prev_option()
        elif key == curses.KEY_DOWN:
            self.next_option()
        elif key == 27:
            self._enable = False
        elif key == InputManager.backspace1 or key == InputManager.backspace2 or key == InputManager.delete:
            self.search_text = self.search_text[:-1]
            self.update_index()
        elif key >= 32 and key < 127:
            self.search_text += chr(key).lower()
            self.update_index()
        elif key == ord('\n'):
            if self._exit_on_action or self._options[self._index].exit_on_action:
                self._enable = False
            self._options[self._index].action()
            return -1
        else:
            return key
        return -1



from .text import Token

class __Symbols:

    def __init__(self):
        self.downloaded = Token("▼")
        self.to_download = Token("▽")
        self.cant_download = Token("◉")
        self.opening = Token("=> ")
        self.neutral = Token("»")
        self.success = Token("✓")
        self.failure = Token("✗")
        self.wrong = Token("ω")
        self.compilation = Token("ϲ")
        self.execution = Token("ϵ")
        self.unequal = Token("├")
        self.equalbar = Token("│")
        self.hbar = Token("─")
        self.vbar = Token("│")

        self.whitespace = Token("·")
        # self.whitespace = Token("␣")
        
        # self.newline = Token("¶")
        self.newline = Token("↲")
        # self.newline = Token("⏎")

        self.cfill = Token("_")
        self.tab = Token("    ")
        self.arrow_up = Token("↑")

        self.check = Token("✓")
        self.uncheck = Token("✗")
        # self.opcheck = Token("ⴲ")
        # self.opuncheck = Token("ⵔ")
        self.infinity = Token("∞")
        self.locked_free = Token("⇉")
        self.locked_locked = Token("⇟")
        self.left_toggle = Token("━─")
        self.right_toggle = Token("─━")
        self.timer = Token("⏰") # ⌛⏰
        self.diff_down = Token("⇕")
        self.diff_left = Token("⇔")
        self.sharpR = Token("")
        self.sharpL = Token("")
        self.action = Token("◎", "b")

        self.emoji_vazio = Token("✗", "r") #"❌"
        self.emoji_confiante = Token("●", "g")#"😎"
        self.emoji_capaz     = Token("◕", "y")#"😌"
        self.emoji_inseguro  = Token("◑", "m")#"😬"
        self.emoji_confuso   = Token("◔", "r")#"😕"

        self.emoji_alone = Token("A", "g")#"💖" # "✋"
        self.emoji_dicas = Token("B", "y") # "💡"
        self.emoji_codes = Token("C", "m")#"💚" #"🤝"
        self.emoji_guide = Token("D", "r")#"💙" # "🧭"
        self.cursor = Token("┊")

    def set_colors(self):
        self.opening.fmt = "b"
        self.neutral.fmt = "b"
        self.success.fmt = "g"
        self.failure.fmt = "r"
        self.wrong.fmt = "r"
        self.compilation.fmt = "y"
        self.execution.fmt = "y"
        self.unequal.fmt = "r"
        self.equalbar.fmt = "g"

symbols = __Symbols()



import unittest
from datetime import datetime
from unittest.mock import MagicMock
from tko.util.logger import Logger, LoggerMemory, LogAction, ActionData

class TestLogger(unittest.TestCase):

    def setUp(self):
        self.logger_store = LoggerMemory()
        self.logger = Logger(self.logger_store)
        self.logger.set_rep("test_rep")

    def test_set_rep(self):
        self.logger.set_rep("new_rep")
        self.assertEqual(self.logger.rep_alias, "new_rep")
        self.assertEqual(self.logger_store.rep_alias, "new_rep")

    def test_record_event(self):
        self.logger.record_event(LogAction.OPEN, "task1", "payload1")
        entries = self.logger_store.get_action_entries()
        self.assertEqual(len(entries), 1)
        self.assertEqual(entries[0].action_value, LogAction.OPEN.value)
        self.assertEqual(entries[0].task_key, "task1")
        self.assertEqual(entries[0].payload, "payload1")

    def test_get_last_hash(self):
        self.logger.record_event(LogAction.OPEN, "task1", "payload1")
        last_hash = self.logger.get_last_hash()
        self.assertEqual(last_hash, self.logger_store.entries[0].hash)

    def test_check_log_file_integrity(self):
        self.logger.record_event(LogAction.OPEN, "task1", "payload1")
        self.logger.record_event(LogAction.QUIT, "task2", "payload2")
        integrity_issues = self.logger.check_log_file_integrity()
        self.assertEqual(len(integrity_issues), 0)

    def test_store_in_cached(self):
        action_data = ActionData(datetime.now().strftime('%Y-%m-%d %H:%M:%S'), LogAction.SELF.value, "test_rep", "task1", "payload1")
        self.assertTrue(self.logger.store_in_cached(action_data))
        self.logger.cached_action = action_data
        self.assertTrue(self.logger.store_in_cached(action_data))

    def test_record_action_data(self):
        action_data = ActionData("um", LogAction.SELF.value, "test_rep", "task1", "payload1")
        self.logger.record_action_data(action_data)
        action_data = ActionData("dois", LogAction.SELF.value, "test_rep", "task1", "payload2")
        self.logger.record_action_data(action_data)
        self.assertEqual(len(self.logger_store.entries), 0)
        action_data = ActionData("tres", LogAction.SELF.value, "test_rep", "task2", "payload1")
        self.logger.record_action_data(action_data)
        self.assertEqual(len(self.logger_store.entries), 1)
        action_data = ActionData("quatro", LogAction.QUIT.value, "test_rep")
        self.logger.record_action_data(action_data)
        self.assertEqual(len(self.logger_store.entries), 3)
        self.assertEqual(self.logger_store.entries[0].task_key, "task1")
        self.assertEqual(self.logger_store.entries[1].task_key, "task2")
        self.assertEqual(self.logger_store.entries[2].action_value, LogAction.QUIT.value)

    def test_check_log_file_integrity2(self):
        self.logger.record_event(LogAction.OPEN, "task1", "payload1")
        self.logger.record_event(LogAction.QUIT, "task2", "payload2")
        self.logger.record_event(LogAction.QUIT, "task3", "payload3")
        integrity_issues = self.logger.check_log_file_integrity()
        self.assertEqual(len(integrity_issues), 0)
        self.logger_store.entries[1].hash = "wrong_hash"
        integrity_issues = self.logger.check_log_file_integrity()
        self.assertEqual(len(integrity_issues), 1)

if __name__ == '__main__':
    unittest.main()


from typing import Dict
from tko.util.text import Text, Token
from tko.util.symbols import symbols

class GradeMessage:
    def __init__(self):
        self.msg = r"""
  Não Fiz  │Compreensão→│Superficial│    Básico │Funcional │   Profundo 
↓Autonomia │            │$1 Confuso  │$2 Inseguro │$3 Capaz   │$4 Confiante
───────────┼────────────┼───────────┼───────────┼──────────┼────────────
Sem   Ajuda│$5  Sozinho  │           │           │   [8]    │    [10]    
Pouca Ajuda│$6  Dicas    │           │    [4]    │   [7]    │    [9]     
Muita Ajuda│$7  Códigos  │           │    [3]    │   [6]    │            
Tutorial   │$8  Guiado   │    [1]    │    [2]    │   [5]    │            
"""[1:-1]
        self.emoji: Dict[str, Token] = {
            "$1": symbols.emoji_confuso,
            "$2": symbols.emoji_inseguro,
            "$3": symbols.emoji_capaz,
            "$4": symbols.emoji_confiante,
            "$5": symbols.emoji_alone,
            "$6": symbols.emoji_dicas,
            "$7": symbols.emoji_codes,
            "$8": symbols.emoji_guide
        }

        self.axes: Dict[int, list[str]] = {}
        self.load_axes()

    @staticmethod
    def grade_to_emojis(grade: int) -> Text:
        decode_dict = {
                0: (Token("-"), Token("-", " ")),
                1: (symbols.emoji_guide, symbols.emoji_confuso),
                2: (symbols.emoji_guide, symbols.emoji_inseguro),
                3: (symbols.emoji_codes, symbols.emoji_inseguro),
                4: (symbols.emoji_dicas, symbols.emoji_inseguro),
                5: (symbols.emoji_guide, symbols.emoji_capaz),
                6: (symbols.emoji_codes, symbols.emoji_capaz),
                7: (symbols.emoji_dicas, symbols.emoji_capaz),
                8: (symbols.emoji_alone, symbols.emoji_capaz),
                9: (symbols.emoji_dicas, symbols.emoji_confiante),
                10: (symbols.emoji_alone, symbols.emoji_confiante)
        }
        a = decode_dict[grade][0]
        b = decode_dict[grade][1]
        return Text("{} {}", b, a)

    def load_axes(self):
            sozinho = "Sozinho"
            dicas   = "Dicas  "
            codigos = "Códigos"
            guiado  = "Guiado "
            
            superficial = "Superficial"
            basico = "    Básico "
            funcional = "Funcional "
            profundo = "   Profundo "

            sem_ajuda   = "Sem   Ajuda"
            pouca_ajuda = "Pouca Ajuda"
            muita_ajuda = "Muita Ajuda"
            tutorial    = "Tutorial   "

            confuso   = "Confuso "
            inseguro  = "Inseguro"
            capaz     = "Capaz  "
            confiante = "Confiante"

            self.axes = {
                1: [tutorial, guiado, superficial, confuso],
                2: [tutorial, guiado, basico, inseguro],
                3: [muita_ajuda, codigos, basico, inseguro],
                4: [pouca_ajuda, dicas, basico, inseguro],
                5: [tutorial, guiado, funcional, capaz],
                6: [muita_ajuda, codigos, funcional, capaz],
                7: [pouca_ajuda, dicas, funcional, capaz],
                8: [sem_ajuda, sozinho, funcional, capaz],
                9: [pouca_ajuda, dicas, profundo, confiante],
                10: [sem_ajuda, sozinho, profundo, confiante]
            }

    def format(self, grade: int):
            headers = [Token("Compreensão→", "B"), Token("↓Autonomia ", "M")]

            value: Text = Text(self.msg)
            if grade != 0:
                for h in headers:
                    value = value.replace(h.text, h)
            info = str(grade)
            pos = value.get_text().find(" [{}] ".format(info))
            if pos != -1:
                while True:
                    if pos >= len(value) or value.data[pos].text == "│" or value.data[pos].text == "\n":
                        break
                    value.data[pos].fmt = "G"
                    pos += 1
            if grade > 0 and grade <= 10:
                value = value.replace(self.axes[grade][0], Token(self.axes[grade][0], "M"))
                value = value.replace(self.axes[grade][1], Token(self.axes[grade][1], "g"))
                value = value.replace(self.axes[grade][2], Token(self.axes[grade][2], "B"))
                value = value.replace(self.axes[grade][3], Token(self.axes[grade][3], "g"))

                value = value.replace("  Não Fiz  ", Token("           "))
            elif grade == 0:
                value = value.replace("  Não Fiz  ", Token("  Não Fiz  ", "R"))

            for k, v in self.emoji.items():
                value = value.replace(k, v)

            return value
    



from __future__ import annotations
import csv
import hashlib
import datetime
from tko.settings.settings import Settings
import enum
import os

from abc import ABC, abstractmethod


class LogAction(enum.Enum):
    NONE = 'NONE'
    OPEN = 'OPEN'
    DOWN = 'DOWN'
    FREE = 'FREE' 
    TEST = 'TEST' # CORRECT|WRONG|COMPILE|RUNTIME
    SELF = 'SELF' # int
    QUIT = 'QUIT'


class ActionData:
    def __init__(self, timestamp: str, action_value: str, rep: str = "", task: str = "", payload: str = ""):
        self.timestamp = timestamp
        self.action_value: str = action_value
        self.rep_alias = rep
        self.task_key = task
        self.payload = payload
        self.hash = ""

    def action_self_same_task(self, other: ActionData):
        if self.action_value != LogAction.SELF.value:
            return False
        if other.action_value != LogAction.SELF.value:
            return False
        return self.rep_alias == other.rep_alias and self.task_key == other.task_key

    def __str__(self):
        return f'{self.timestamp}, {self.action_value}, {self.task_key}, {self.payload}'

    @staticmethod
    def generate_hash(action_data: ActionData, last_hash: str):
        comprimento = 6
        input_str = str(action_data) + last_hash
        hash_completo = hashlib.sha256(input_str.encode()).hexdigest()
        return hash_completo[:comprimento]  # Retorna os primeiros 'comprimento' caracteres do hash

class LoggerStore(ABC):
    @abstractmethod
    def push_to_file(self, action_data: ActionData, last_hash: str):
        pass

    @abstractmethod
    def get_action_entries(self) -> list[ActionData]:
        pass

    @abstractmethod
    def set_rep(self, rep_alias: str) -> None:
        pass

    @abstractmethod
    def get_last_hash(self) -> str:
        pass

class LoggerFS(LoggerStore):

    def __init__(self, settings: Settings):
        self.log_file: str | None = None
        self.settings = settings
        self.rep_alias = ""

    def set_rep(self, rep_alias: str):
        self.rep_alias = rep_alias
        return self

    def set_log_file(self, log_file: str):
        self.log_file = log_file
        return self
    
    def get_log_file(self) -> str:
        if self.log_file is not None:
            return self.log_file
        rootdir = self.settings.app.get_rootdir()
        log_file = "rep_log.csv"
        return os.path.join(rootdir, self.rep_alias, log_file)
    
    def row_to_action_data(self, row: list[str]) -> ActionData:
        hash = row[0]
        timestamp = row[1]
        action_value = row[2]
        task = row[3]
        payload = row[4]
        action_data = ActionData(timestamp, action_value, self.rep_alias, task, payload)
        action_data.hash = hash
        return action_data

    def push_to_file(self, action_data: ActionData, last_hash: str):
        if self.rep_alias == "":
            return
        action_data.hash = ActionData.generate_hash(action_data, last_hash)
        
        with open(self.get_log_file(), 'a', newline='') as file:
            writer = csv.writer(file)
            ad = action_data
            writer.writerow([ad.hash, ad.timestamp, ad.action_value, ad.task_key, ad.payload])
            return ad.hash

    def get_action_entries(self) -> list[ActionData]:
        log_file = self.get_log_file()
        if not os.path.exists(log_file):
            return []
        with open(log_file, 'r') as file:
            reader = csv.reader(file)
            rows = list(reader)
            return [self.row_to_action_data(row) for row in rows]

    def get_last_hash(self) -> str:
        return self.get_action_entries()[-1].hash if len(self.get_action_entries()) > 0 else ""

class LoggerMemory(LoggerStore):
    def __init__(self):
        self.entries: list[ActionData] = []
        self.rep_alias = ""

    def set_rep(self, rep_alias: str):
        self.rep_alias = rep_alias
        return self

    def push_to_file(self, action_data: ActionData, last_hash: str):
        action_data.hash = ActionData.generate_hash(action_data, last_hash)
        self.entries.append(action_data)
        return action_data.hash

    def get_action_entries(self) -> list[ActionData]:
        return self.entries

    def get_last_hash(self) -> str:
        return self.entries[-1].hash if len(self.entries) > 0 else ""

class Logger:
    instance: None | Logger = None
    COMP_ERROR = "COMPILATION_ERROR"

    @staticmethod
    def get_instance() -> Logger:
        if Logger.instance is None:
            raise Exception("Logger not initialized")
        return Logger.instance

    def __init__(self, logger_store: LoggerStore):
        self.last_hash: str | None = None
        self.rep_alias: str = ""
        self.cached_action: ActionData | None = None
        self.fs = logger_store

    def set_rep(self, rep_alias: str):
        self.rep_alias = rep_alias
        self.fs.set_rep(rep_alias)
        return self

    def check_log_file_integrity(self) -> list[str]:
        entries = self.fs.get_action_entries()
        if len(entries) == 0:
            return []
        hash = entries[0].hash
        output: list[str] = []

        for i in range(1, len(entries)):
            calculated_hash = ActionData.generate_hash(entries[i], hash)
            if calculated_hash != entries[i].hash:
                output.append(f"Hash mismatch line {i + 1}: {str(entries[i])}")
            hash = calculated_hash
        return output

    def store_in_cached(self, action_data: ActionData) -> bool:
        if self.cached_action is None and action_data.action_value == LogAction.SELF.value:
            return True
        if self.cached_action is not None and self.cached_action.action_self_same_task(action_data):
            return True
        return False

    def record_event(self, action: LogAction, task_key: str = "", payload: str = ""):
        if self.rep_alias == "":
            return
        action_data = ActionData(datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), action.value, self.rep_alias, task_key, payload)
        self.record_action_data(action_data)

    def get_last_hash(self) -> str:
        if self.last_hash is None:
            self.last_hash = self.fs.get_last_hash()
        return self.last_hash

    def record_action_data(self, action_data: ActionData):
        if self.store_in_cached(action_data):
            self.cached_action = action_data
            return
        
        if self.cached_action is not None:
            self.last_hash = self.fs.push_to_file(self.cached_action, self.get_last_hash())
            self.cached_action = None
        
        if self.store_in_cached(action_data):
            self.cached_action = action_data
        else:
            self.last_hash = self.fs.push_to_file(action_data, self.get_last_hash())



from typing import Optional, Union
import shutil
from .text import Text, Token


class RawTerminal:
    __term_width: Optional[int] = None

    def __init__(self):
        pass

    @staticmethod
    def __get_terminal_size() -> int:
        return shutil.get_terminal_size().columns

    @staticmethod
    def get_terminal_size():
        if RawTerminal.__term_width is None:
            return RawTerminal.__get_terminal_size()
        return RawTerminal.__term_width

    @staticmethod
    def set_terminal_size(value: int):
        RawTerminal.__term_width = value

    @staticmethod
    def centralize(text: Text | str, filler: Token | str = Token(" ")):
        if isinstance(filler, str):
            filler = Token(filler)

        if isinstance(text, str):
            text = Text(text)
        width = RawTerminal.get_terminal_size()
        return text.center(width, filler)
    


from tko.run.diff_builder import DiffBuilder
from tko.run.unit import Unit
from tko.util.consts import ExecutionResult
from tko.util.symbols import symbols
from tko.util.text import Text, Token


from typing import List, Tuple


class DownDiff:
    def __init__(self, width: int, unit: Unit):
        self.width = width
        self.curses = False
        self.db = DiffBuilder(width)
        self.unit = unit
        self.output: List[Text] = []
        self.__to_insert_header = False
        self.no_diff_mode = self.unit.input == "" and self.unit.expected == ""
        self.expected_received, self.first_failure = self.db.render_diff(self.unit.expected, self.unit.received)

    def to_insert_header(self):
        self.__to_insert_header = True
        return self

    def set_curses(self):
        self.curses = True
        return self

    @staticmethod
    def put_left_equal(exp_rec_list: List[Tuple[Text | None, Text | None]], unequal: Token = symbols.unequal):

        output: List[Tuple[Text, Text]] = []
        for exp, rec in exp_rec_list:
            if exp is None or rec is None or exp != rec:
                exp_lines = Text() + unequal + " " + exp
                rec_lines = Text() + unequal + " " + rec
            else:
                exp_lines = Text() + symbols.vbar + " " + exp
                rec_lines = Text() + symbols.vbar + " " + rec
            output.append((exp_lines, rec_lines))
        return output

    def insert_header(self):
        self.output.append(Text("").fold_in(self.width, symbols.hbar, "╭", "╮"))
        self.output.append(self.unit.str().fold_in(self.width, " ", symbols.vbar, symbols.vbar))

    def insert_input(self):
        color = "g" if self.unit.expected == self.unit.received else "b"
        # header
        if self.__to_insert_header:
            self.output.append(Text().addf(color, DiffBuilder.vinput).fold_in(self.width, symbols.hbar, "├", "┤"))
        else:
            self.output.append(Text().addf(color, DiffBuilder.vinput).fold_in(self.width, symbols.hbar, "╭", "╮"))
        # lines
        for line in self.unit.input.split("\n")[:-1]:
            self.output.append(Text().add(symbols.vbar).add(" ").add(line).ljust(self.width - 1, Token(" ")).add(symbols.vbar))

    def insert_expected(self):
        if self.no_diff_mode:
            return
        # color = "g" if self.unit.expected == self.unit.received else "b"
        # self.output.append(Text().addf(color, DiffBuilder.vreceived).fold_in(self.width, symbols.hbar, "╭", "╮"))
        if self.unit.expected == "":
            return
        self.output.append(Text().addf("g", DiffBuilder.vexpected).fold_in(self.width, symbols.hbar, "├", "┤"))
        for line, _ in self.expected_received:
            if line is not None:
                self.output.append(line.ljust(self.width - 1, Token(" ")).add(symbols.vbar))

    def insert_received(self):
        # headers
        color = "r" if self.unit.expected == self.unit.received else "g"
        if self.no_diff_mode:
            color = "g"
            self.output.append(Text().addf(color, DiffBuilder.vreceived).fold_in(self.width, symbols.hbar, "╭", "╮"))
        else:
            self.output.append(Text().addf(color, DiffBuilder.vreceived).fold_in(self.width, symbols.hbar, "├", "┤"))

        # lines
        for _, line in self.expected_received:
            if line is not None:
                self.output.append(line.ljust(self.width - 1, Token(" ")).trim_end(self.width).add(symbols.vbar))

    def insert_first_line_diff(self):
        include_rendering = False
        if self.unit.expected != self.unit.received and self.unit.expected != "":
            include_rendering = True
        if self.unit.result == ExecutionResult.EXECUTION_ERROR or self.unit.result == ExecutionResult.COMPILATION_ERROR:
            include_rendering = False

        if not include_rendering:
            return
        self.output.append(Text().addf("b", DiffBuilder.vunequal).fold_in(self.width, symbols.hbar, "├", "┤"))
        for line in self.db.first_failure_diff(self.unit.expected, self.unit.received, self.first_failure):
            self.output.append(Text("│").add(line).ljust(self.width - 1, Token(" ")).add("│"))

    def end_frame(self):
        self.output.append(Text("").fold_in(self.width, symbols.hbar, "╰", "╯"))

    def build_diff(self) -> List[Text]:
        if self.__to_insert_header:
            self.insert_header()
        if not self.no_diff_mode:
            self.insert_input()
        symb = symbols.unequal
        if self.unit.result == ExecutionResult.EXECUTION_ERROR or self.unit.result == ExecutionResult.COMPILATION_ERROR or self.unit.expected == "":
            symb = symbols.vbar
        self.expected_received = self.put_left_equal(self.expected_received, symb)
        self.insert_expected()
        self.insert_received()
        self.insert_first_line_diff()
        self.end_frame()

        return self.output


from tko.game.game import Game
from tko.game.graph import Graph
from tko.settings.rep_settings import RepSource
from tko.settings.settings import Settings
from tko.util.logger import Logger

import os

class CmdRep:
    @staticmethod
    def check(args):
        rep = args.alias
        logger = Logger.get_instance()
        logger.set_rep(rep)
        output = logger.check_log_file_integrity()
        if len(output) == 0:
            print(f"Arquivo de log do repositório {rep} está íntegro.")
        else:
            print(f"Arquivo de log do repositório {rep} está corrompido.")
            print("Erros:")
            for error in output:
                print(f"- {error}")

    @staticmethod
    def list(_args):
        settings = Settings()
        print(f"SettingsFile\n- {settings.settings_file}")
        print(str(settings))

    @staticmethod
    def add(args):
        settings = Settings()
        rep = RepSource()
        if args.url:
            rep.set_url(args.url)
        elif args.file:
            rep.set_file(args.file)
        settings.reps[args.alias] = rep
        settings.save_settings()

    @staticmethod
    def rm(args):
        sp = Settings()
        if args.alias in sp.reps:
            sp.reps.pop(args.alias)
            sp.save_settings()
        else:
            print("Repository not found.")

    @staticmethod
    def reset(_):
        sp = Settings().reset()
        print(sp.settings_file)
        print(sp.app._rootdir)
        sp.save_settings()

    @staticmethod
    def graph(args):
        settings = Settings()
        rep_source:RepSource = settings.get_rep_source(args.alias)
        file = rep_source.get_file_or_cache(os.path.join(settings.app._rootdir, args.alias))
        game = Game()
        game.parse_file(file)
        game.check_cycle()
        Graph(game).generate()


import unittest
from .compare import compare_list

class Test:
    def test_simple_list1(self, capsys):
        folder = "tests/02_list"
        compare_list(capsys, folder, "out1", ["-m", "run", "cases.tio"])

    def test_simple_list2(self, capsys):
        folder = "tests/02_list"
        compare_list(capsys, folder, "out2", ["-m", "run", "cases.tio", "cases2.tio"])

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import argparse
import sys

from tko.cmds.cmd_rep import CmdRep
from tko.cmds.cmd_play import CmdPlay
from tko.cmds.cmd_down import CmdDown
from tko.cmds.cmd_run import Run
from tko.cmds.cmd_build import CmdBuild
from tko.cmds.cmd_config import CmdConfig, ConfigParams
from tko.util.logger import LogAction, Logger, LoggerFS
from tko.util.text import AnsiColor

from tko.util.param import Param
from tko.util.pattern import PatternLoader
from tko.util.consts import DiffCount, DiffMode
from .settings.settings import Settings

from tko.util.guide import tko_guide
from tko.util.guide import bash_guide

from tko.util.raw_terminal import RawTerminal
from tko.util.symbols import symbols
from tko.settings.check_version import CheckVersion

from .__init__ import __version__


class Main:
    @staticmethod
    def test(args):
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        if args.quiet:
            param.set_diff_count(DiffCount.QUIET)
        elif args.all:
            param.set_diff_count(DiffCount.ALL)
        else:
            param.set_diff_count(DiffCount.FIRST)

        if args.filter:
            param.set_filter(True)
        if args.compact:
            param.set_compact(True)

        settings = Settings()
        if not args.side and not args.down:
            param.set_diff_mode(settings.app.get_diff_mode())
        elif args.side:
            param.set_diff_mode(DiffMode.SIDE)
        elif args.down:
            param.set_diff_mode(DiffMode.DOWN)
        cmd_run = Run(settings, args.target_list, args.cmd, param)
        cmd_run.execute()

    @staticmethod
    def run(args):
        PatternLoader.pattern = args.pattern
        param = Param.Basic().set_index(args.index)
        settings = Settings()
        param.set_diff_mode(settings.app.get_diff_mode())
        if args.filter:
            param.set_filter(True)
        cmd_run = Run(settings, args.target_list, args.cmd, param)
        cmd_run.set_curses()
        cmd_run.execute()

    @staticmethod
    def build(args):
        PatternLoader.pattern = args.pattern
        manip = Param.Manip().set_unlabel(args.unlabel).set_to_sort(args.sort).set_to_number(args.number)
        build = CmdBuild(args.target, args.target_list, manip, args.force)
        build.execute()

    @staticmethod
    def play(args):
        settings = Settings()
        settings.check_rootdir()
        CmdPlay.execute(args.repo, settings)
        CheckVersion().version_check()

    @staticmethod
    def down(args):
        settings = Settings().check_rootdir()
        CmdDown.execute(args.course, args.activity, args.language, settings, print)

    @staticmethod
    def config(args):
        settings = Settings()
        param = ConfigParams()
        param.side = args.side
        param.down = args.down
        param.lang = args.lang
        param.ask = args.ask
        param.root = args.root
        param.editor = args.editor
        param.borders = args.borders
        param.hud = args.hud

        CmdConfig.execute(settings, param)


class Parser:
    def __init__(self):
        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(prog='tko', description=f'tko version {__version__}')
        self.subparsers = self.parser.add_subparsers(title='subcommands', help='help for subcommand.')

        self.parent_manip = self.create_parent_manip()
        self.parent_basic = self.create_parent_basic()

        self.add_parser_global()
        self.create_parent_basic()
        self.create_parent_manip()
        self.add_parser_run()
        self.add_parser_exec()
        self.add_parser_build()
        self.add_parser_down()
        self.add_parser_config()
        self.add_parser_repo()
        self.add_parser_play()

    def add_parser_global(self):
        self.parser.add_argument('-c', metavar='CONFIG_FILE', type=str, help='config json file.')
        self.parser.add_argument('-w', metavar='WIDTH', type=int, help="terminal width.")
        self.parser.add_argument('-v', action='store_true', help='show version.')
        self.parser.add_argument('-m', action='store_true', help='monochromatic.')

    def create_parent_basic(self):
        parent_basic = argparse.ArgumentParser(add_help=False)
        parent_basic.add_argument('--index', '-i', metavar="I", type=int, help='run a specific index.')
        parent_basic.add_argument('--pattern', '-p', metavar="P", type=str, default='@.in @.sol',
                                  help='pattern load/save a folder, default: "@.in @.sol"')
        return parent_basic

    def create_parent_manip(self):
        parent_manip = argparse.ArgumentParser(add_help=False)
        parent_manip.add_argument('--width', '-w', type=int, help="term width.")
        parent_manip.add_argument('--unlabel', '-u', action='store_true', help='remove all labels.')
        parent_manip.add_argument('--number', '-n', action='store_true', help='number labels.')
        parent_manip.add_argument('--sort', '-s', action='store_true', help="sort test cases by input size.")
        parent_manip.add_argument('--pattern', '-p', metavar="@.in @.out", type=str, default='@.in @.sol',
                                  help='pattern load/save a folder, default: "@.in @.sol"')
        return parent_manip

    def add_parser_run(self):
        parser_r = self.subparsers.add_parser('run', parents=[self.parent_basic], help='run with test cases using curses.')
        parser_r.add_argument('target_list', metavar='T', type=str, nargs='*', help='solvers, test cases or folders.')
        parser_r.add_argument('--filter', '-f', action='store_true', help='filter solver in temp dir before run')
        parser_r.add_argument("--cmd", type=str, help="bash command to run code")
        parser_r.set_defaults(func=Main.run)

    def add_parser_exec(self):
        parser_r = self.subparsers.add_parser('test', parents=[self.parent_basic], help='run with test cases.')
        parser_r.add_argument('target_list', metavar='T', type=str, nargs='*', help='solvers, test cases or folders.')
        parser_r.add_argument('--filter', '-f', action='store_true', help='filter solver in temp dir before run')
        parser_r.add_argument('--compact', '-c', action='store_true', help='Do not show case descriptions in failures')
        parser_r.add_argument("--cmd", type=str, help="bash command to run code")

        group_n = parser_r.add_mutually_exclusive_group()
        group_n.add_argument('--quiet', '-q', action='store_true', help='quiet mode, do not show any failure.')
        group_n.add_argument('--all', '-a', action='store_true', help='show all failures.')

        # add an exclusive group for diff mode
        group = parser_r.add_mutually_exclusive_group()
        group.add_argument('--down', '-d', action='store_true', help="diff mode up-to-down.")
        group.add_argument('--side', '-s', action='store_true', help="diff mode side-by-side.")
        parser_r.set_defaults(func=Main.test)

    def add_parser_build(self):
        parser_b = self.subparsers.add_parser('build', parents=[self.parent_manip], help='build a test target.')
        parser_b.add_argument('target', metavar='T_OUT', type=str, help='target to be build.')
        parser_b.add_argument('target_list', metavar='T', type=str, nargs='+', help='input test targets.')
        parser_b.add_argument('--force', '-f', action='store_true', help='enable overwrite.')
        parser_b.set_defaults(func=Main.build)

    def add_parser_down(self):
        parser_d = self.subparsers.add_parser('down', help='download problem from repository.')
        parser_d.add_argument('course', type=str, nargs='?', help=" [ fup | ed | poo ].")
        parser_d.add_argument('activity', type=str, nargs='?', help="activity @label.")
        parser_d.add_argument('--language', '-l', type=str, nargs='?', help="[ c | cpp | js | ts | py | java ]")
        parser_d.set_defaults(func=Main.down)

    def add_parser_config(self):
        parser_s = self.subparsers.add_parser('config', help='settings tool.')

        g_diff = parser_s.add_mutually_exclusive_group()
        g_diff.add_argument('--side', action='store_true', help='set side_by_side diff mode.')
        g_diff.add_argument('--down', action='store_true', help='set up_to_down   diff mode.')

        g_lang = parser_s.add_mutually_exclusive_group()
        g_lang.add_argument("--lang", '-l', metavar='ext', type=str, help="set default language extension.")
        g_lang.add_argument("--ask", action='store_true', help='ask language extension every time.')

        parser_s.add_argument("--root", metavar="path", type=str, help='set root directory.')
        parser_s.add_argument("--editor", metavar="cmd", type=str, help='set editor command.')
        parser_s.add_argument("--borders", metavar="0or1", type=str, help='enable borders.')
        parser_s.add_argument("--hud", metavar="0or1", type=str, help='enable hud.')

        parser_s.set_defaults(func=Main.config)

    def add_parser_repo(self):
        parser_repo = self.subparsers.add_parser('rep', help='manipulate repositories.')
        subpar_repo = parser_repo.add_subparsers(title='subcommands', help='help for subcommand.')

        repo_list = subpar_repo.add_parser('list', help='list all repositories')
        repo_list.set_defaults(func=CmdRep.list)

        repo_add = subpar_repo.add_parser('add', help='add a repository.')
        repo_add.add_argument('alias', metavar='alias', type=str, help='alias of the repository to be added.')
        repo_add.add_argument('--url', '-u', type=str, help='add a repository url to the settings file.')
        repo_add.add_argument('--file', '-f', type=str, help='add a repository file to the settings file.')
        repo_add.set_defaults(func=CmdRep.add)

        repo_rm = subpar_repo.add_parser('rm', help='remove a repository.')
        repo_rm.add_argument('alias', metavar='alias', type=str, help='alias of the repository to be removed.')
        repo_rm.set_defaults(func=CmdRep.rm)

        repo_reset = subpar_repo.add_parser('reset', help='reset all repositories to factory default.')
        repo_reset.set_defaults(func=CmdRep.reset)

        repo_graph = subpar_repo.add_parser('graph', help='generate graph of the repository.')
        repo_graph.add_argument('alias', metavar='alias', type=str, help='alias of the repository to be graphed.')
        repo_graph.set_defaults(func=CmdRep.graph)

        repo_log = subpar_repo.add_parser("check", help="validates log of the repository.")
        repo_log.add_argument('alias', metavar='alias', type=str, help='alias of the repository to be log.')
        repo_log.set_defaults(func=CmdRep.check)

    def add_parser_play(self):
        parser_p = self.subparsers.add_parser('play', help='play a game.')
        parser_p.add_argument('repo', metavar='repo', type=str, nargs="?", default="__ask", help='repository to be played.')
        # parser_p.add_argument("--graph", "-g", action='store_true', help='generate graph of the game using graphviz.')
        # parser_p.add_argument("--svg", "-s", action='store_true', help='generate graph in svg instead png.')
        parser_p.set_defaults(func=Main.play)


def exec(parser: argparse.ArgumentParser, args):
    settings = Settings()
    if args.w is not None:
        RawTerminal.set_terminal_size(args.w)
    if args.c:
        settings.set_settings_file(args.c)
    settings.load_settings()
    Logger.instance = Logger(LoggerFS(settings))

    if args.m:
        AnsiColor.enabled = False
    else:
        AnsiColor.enabled = True
        symbols.set_colors()

    if args.v:
        if args.v:
            print("tko version " + __version__)
    else:
        if "func" in args:
            args.func(args)
        else:
            parser.print_help()

def main():
    try:
        parser = Parser().parser
        args = parser.parse_args()
        exec(parser, args)
        sys.exit(0)
    except KeyboardInterrupt:
        print("\n\nKeyboard Interrupt")
        sys.exit(1)
    except Warning as w:
        print(w)
        sys.exit(1)
    # except Exception as e:
    #     print(e)
    #     sys.exit(1)

# NÃO RETIRE ESSA TAG POR CAUSA DO MERGE DO REPLIT
if __name__ == '__main__': # MERGE_INSERT 
    main()



from typing import List, Dict, Any
from ..util.text import Text


class Flag:
    def __init__(self):
        self._name: str = ""
        self._text: str = ""  # description
        self._char: str = ""
        self._values: List[str] = ["0", "1"]
        self._index: int = 0
        self._location: str = ""
        self._bool = True  # many options

    def set_name(self, _name):
        self._name = _name
        return self

    def set_description(self, _text):
        self._text = _text
        return self

    def set_keycode(self, _key):
        self._char = _key
        return self

    def set_values(self, _values: List[str]):
        self._values = _values
        return self

    def index(self, _index):
        self._index = _index
        return self

    def location(self, value: str):
        self._location = value
        return self

    def get_values(self):
        return self._values

    def toggle(self):
        self._index = (self._index + 1) % len(self._values)
        return self
    
    def set_bool(self):
        self._bool = True
        return self
    
    def is_bool(self):
        return self._bool

    def get_value(self) -> str:
        return self._values[self._index % len(self._values)]
    
    def set_value(self, value: Any):
        for i, v in enumerate(self._values):
            if v == value:
                self._index = i
                break

    def is_true(self):
        return self.get_value() == "1"

    def get_name(self) -> str:
        return self._name

    def get_description(self) -> str:
        return self._text

    def get_keycode(self) -> str:
        return self._char

    def get_index(self) -> int:
        return self._index
    

class Flags:
    minimum = Flag().set_name("Mínimo").set_keycode("M").set_values(["0", "1"])    .set_description("Mostra os requisitos para completar a missão")
    reward = Flag().set_name("Recompensa").set_keycode("R").set_values(["0", "1"]) .set_description("Mostra a experiência obtida nas tarefas     ")
    percent = Flag().set_name("Percentual").set_keycode("P").set_values(["1", "0"]).set_description("Mostra todos os valores em porcentagem      ")
    admin = Flag().set_name("Admin").set_keycode("A").set_values(["0", "1"])       .set_description("Habilitas todas as missões e tarefas        ")
    skills = Flag().set_name("Skills").set_keycode("S").set_values(["0", "1"]).set_description("Mostra a barra de skills")
    flags = Flag().set_name("Flags").set_keycode("F").set_values(["0", "1"])       .set_description("Mostra a barra de Flags ")

class FlagsMan:
    def __init__(self, data: Dict[str, int]):
        self.flags: Dict[str, Flag] = {}

        for varname, flag in Flags.__dict__.items():
            if isinstance(flag, Flag):
                self.flags[varname] = flag

        for key, _index in data.items():
            if key in self.flags:
                self.flags[key].index(_index)

    def get_data(self) -> Dict[str, int]:
        data = {}
        for name, flag in self.flags.items():
            if len(flag.get_values()) > 1:
                data[name] = flag.get_index()
        return data



from tko.play.flags import Flag, Flags, FlagsMan
from tko.play.floating import Floating, FloatingInput, FloatingInputData
from tko.play.floating_manager import FloatingManager
from tko.play.border import Border
from tko.settings.settings import Settings
from tko.util.text import Text
from tko.play.border import Border
from typing import List, Tuple, Union, Callable
from tko.play.functors import FlagFunctor
from tko.settings.rep_settings import RepData, languages_avaliable


class LanguageSetter:
    def __init__(self, rep: RepData, flagsman: FlagsMan, fman: FloatingManager):
        self.flagsman = flagsman
        self.fman = fman
        self.rep = rep


    def set_language(self):
        options: List[FloatingInputData] = []
        for lang in languages_avaliable:
            options.append(FloatingInputData(StrFunctor(lang), SetLangFunctor(self.rep, self.fman, lang)))

        self.fman.add_input(
            FloatingInput("^")
            .set_header(" Escolha a extensão default para os rascunhos ")
            .set_options(options)
            .set_default_index(languages_avaliable.index(self.rep.get_lang()))
            .set_footer(" Pressione Enter para confirmar ")
        )

class StrFunctor:
    def __init__(self, data: str):
        self.value = data
    
    def __call__(self):
        return self.value

class SetLangFunctor:
    def __init__(self, rep: RepData, fman: FloatingManager, lang: str):
        self.rep = rep
        self.fman = fman
        self.value = lang

    def __call__(self):
        self.rep.set_lang(self.value.strip())
        self.rep.save_data_to_json()
        self.fman.add_input(
            Floating()
            .put_text("")
            .put_text("Linguagem alterada para " + self.value)
            .put_text("")
            .warning()
        )



from __future__ import annotations
from typing import List, Tuple, Optional
import re
import os

from .unit import Unit
from ..util.pattern import PatternLoader


class VplParser:
    @staticmethod
    def finish(text):
        return text if text.endswith("\n") else text + "\n"

    @staticmethod
    def unwrap(text):
        while text.endswith("\n"):
            text = text[:-1]
        if text.startswith("\"") and text.endswith("\""):
            text = text[1:-1]
        return VplParser.finish(text)

    @staticmethod
    class CaseData:
        def __init__(self, case="", inp="", outp="", grade: Optional[int] = None):
            self.case: str = case
            self.input: str = VplParser.finish(inp)
            self.output: str = VplParser.unwrap(VplParser.finish(outp))
            self.grade: Optional[int] = grade

        def __str__(self):
            return "case=" + self.case + '\n' \
                   + "input=" + self.input \
                   + "output=" + self.output \
                   + "gr=" + str(self.grade)

    regex_vpl_basic = r"case= *([ \S]*) *\n *input *=(.*?)^ *output *=(.*)"
    regex_vpl_extended = r"case= *([ \S]*) *\n *input *=(.*?)^ *output *=(.*?)^ *grade *reduction *= *(\S*)% *\n?"

    @staticmethod
    def filter_quotes(x):
        return x[1:-2] if x.startswith('"') else x

    @staticmethod
    def split_cases(text: str) -> List[str]:
        regex = r"^ *[Cc]ase *="
        subst = "case="
        text = re.sub(regex, subst, text, 0, re.MULTILINE | re.DOTALL)
        return ["case=" + t for t in text.split("case=")][1:]

    @staticmethod
    def extract_extended(text) -> Optional[CaseData]:
        f = re.match(VplParser.regex_vpl_extended, text, re.MULTILINE | re.DOTALL)
        if f is None:
            return None
        try:
            gr = int(f.group(4))
        except ValueError:
            gr = None
        return VplParser.CaseData(f.group(1), f.group(2), f.group(3), gr)

    @staticmethod
    def extract_basic(text) -> Optional[CaseData]:
        m = re.match(VplParser.regex_vpl_basic, text, re.MULTILINE | re.DOTALL)
        if m is None:
            return None
        return VplParser.CaseData(m.group(1), m.group(2), m.group(3), None)

    @staticmethod
    def parse_vpl(content: str) -> List[CaseData]:
        text_cases = VplParser.split_cases(content)
        seq: List[VplParser.CaseData] = []

        for text in text_cases:
            case = VplParser.extract_extended(text)
            if case is not None:
                seq.append(case)
                continue
            case = VplParser.extract_basic(text)
            if case is not None:
                seq.append(case)
                continue
            print("invalid case: " + text)
            exit(1)
        return seq

    @staticmethod
    def to_vpl(unit: CaseData):
        text = "case=" + unit.case + "\n"
        text += "input=" + unit.input
        text += "output=\"" + unit.output + "\"\n"
        if unit.grade is not None:
            text += "grade reduction=" + str(unit.grade) + "%\n"
        return text


class Loader:
    regex_tio = r"^ *>>>>>>>> *(.*?)\n(.*?)^ *======== *\n(.*?)^ *<<<<<<<< *\n?"

    def __init__(self):
        pass

    @staticmethod
    def parse_cio(text, source):
        unit_list = []
        text = "\n" + text

        pattern = r'```.*?\n(.*?)```'  # get only inside code blocks
        code = re.findall(pattern, text, re.MULTILINE | re.DOTALL)
        # join all code blocks found
        text = "\n" + "\n".join(code)

        pieces = []  # header, input, output

        open_case = False
        for line in text.split("\n"):
            if line.startswith("#__case") or line.startswith("#TEST_CASE"):
                pieces.append({"header": line, "input": [], "output": []})
                open_case = True
            elif open_case:
                pieces[-1]["output"].append(line)
                if line.startswith("$end"):
                    open_case = False

        # concatenando testes contínuos e finalizando testes sem $end
        for i in range(len(pieces)):
            output = pieces[i]["output"]
            if output[-1] != "$end" and i < len(pieces) - 1:
                pieces[i + 1]["output"] = output + pieces[i + 1]["output"]
                output.append("$end")

        # removendo linhas vazias e criando input das linhas com $
        for piece in pieces:
            piece["input"] = [line[1:] for line in piece["output"] if line.startswith("$")]
            piece["output"] = [line for line in piece["output"] if line != "" and not line.startswith("#")]

        for piece in pieces:
            case = " ".join(piece["header"].split(" ")[1:])
            inp = "\n".join(piece["input"]) + "\n"
            output = "\n".join(piece["output"]) + "\n"
            unit_list.append(Unit(case, inp, output, None, source))

        return unit_list

    @staticmethod
    def parse_tio(text: str, source: str = "") -> List[Unit]:

        # identifica se tem grade e retorna case name e grade
        def parse_case_grade(value: str) -> Tuple[str, Optional[int]]:
            if value.endswith("%"):
                words = value.split(" ")
                last = value.split(" ")[-1]
                _case = " ".join(words[:-1])
                grade_str = last[:-1]           # ultima palavra sem %
                try:
                    _grade = int(grade_str)
                    return _case, _grade
                except ValueError:
                    pass
            return value, None

        matches = re.findall(Loader.regex_tio, text, re.MULTILINE | re.DOTALL)
        unit_list = []
        for m in matches:
            case, grade = parse_case_grade(m[0])
            unit_list.append(Unit(case, m[1], m[2], grade, source))
        return unit_list

    @staticmethod
    def parse_vpl(text: str, source: str = "") -> List[Unit]:
        data_list = VplParser.parse_vpl(text)
        output: List[Unit] = []
        for m in data_list:
            output.append(Unit(m.case, m.input, m.output, m.grade, source))
        return output

    @staticmethod
    def parse_dir(folder) -> List[Unit]:
        pattern_loader = PatternLoader()
        files = sorted(os.listdir(folder))
        matches = pattern_loader.get_file_sources(files)

        unit_list: List[Unit] = []
        try:
            for m in matches:
                unit = Unit()
                unit.source = os.path.join(folder, m.label)
                unit.grade = 100
                with open(os.path.join(folder, m.input_file)) as f:
                    value = f.read()
                    unit.input = value + ("" if value.endswith("\n") else "\n")
                with open(os.path.join(folder, m.output_file)) as f:
                    value = f.read()
                    unit.expected = value + ("" if value.endswith("\n") else "\n")
                unit_list.append(unit)
        except FileNotFoundError as e:
            print(str(e))
        return unit_list

    @staticmethod
    def parse_source(source: str) -> List[Unit]:
        if os.path.isdir(source):
            return Loader.parse_dir(source)
        if os.path.isfile(source):
            #  if PreScript.exists():
            #      source = PreScript.process_source(source)
            with open(source, encoding="utf-8") as f:
                content = f.read()
            if source.endswith(".vpl"):
                return Loader.parse_vpl(content, source)
            elif source.endswith(".tio"):
                return Loader.parse_tio(content, source)
            elif source.endswith(".md"):
                tests = Loader.parse_tio(content, source)
                tests += Loader.parse_cio(content, source)
                return tests
            else:
                print("warning: target format do not supported: " + source)  # make this a raise
        else:
            raise FileNotFoundError('warning: unable to find: ' + source)
        return []



from typing import List
import os

from ..util.identifier import Identifier

from ..util.pattern import PatternLoader
from ..util.consts import IdentifierType
from .unit import Unit


class Writer:

    def __init__(self):
        pass

    @staticmethod
    def to_vpl(unit: Unit):
        text = "case=" + unit.case + "\n"
        text += "input=" + unit.input
        text += "output=\"" + unit.expected + "\"\n"
        if unit.grade is None:
            text += "\n"
        else:
            text += "grade reduction=" + str(unit.grade).zfill(3) + "%\n"
        return text

    @staticmethod
    def to_tio(unit: Unit):
        text = ">>>>>>>>"
        if unit.case != '':
            text += " " + unit.case
        if unit.grade is not None:
            text += " " + str(unit.grade) + "%"
        text += '\n' + unit.input
        text += "========\n"
        text += unit.expected
        if unit.expected != '' and unit.expected[-1] != '\n':
            text += '\n'
        text += "<<<<<<<<\n"
        return text

    @staticmethod
    def save_dir_files(folder: str, pattern_loader: PatternLoader, label: str, unit: Unit) -> None:
        file_source = pattern_loader.make_file_source(label)
        with open(os.path.join(folder, file_source.input_file), "w") as f:
            f.write(unit.input)
        with open(os.path.join(folder, file_source.output_file), "w") as f:
            f.write(unit.expected)

    @staticmethod
    def save_target(target: str, unit_list: List[Unit], force: bool = False):
        def ask_overwrite(file):
            print("file " + file + " found. Overwrite? (y/n):")
            resp = input()
            if resp.lower() == 'y':
                print("overwrite allowed")
                return True
            print("overwrite denied\n")
            return False

        def save_dir(_target: str, _unit_list):
            folder = _target
            pattern_loader = PatternLoader()
            number = 0
            for unit in _unit_list:
                Writer.save_dir_files(folder, pattern_loader, str(number).zfill(2), unit)
                number += 1

        def save_file(_target, _unit_list):
            if _target.endswith(".tio"):
                _new = "\n".join([Writer.to_tio(unit) for unit in _unit_list])
            else:
                _new = "\n".join([Writer.to_vpl(unit) for unit in _unit_list])

            file_exists = os.path.isfile(_target)

            if file_exists:
                _old = open(_target).read()
                if _old == _new:
                    print("no changes in test file")
                    return

            if not file_exists or (file_exists and (force or ask_overwrite(_target))):
                with open(_target, "w") as f:
                    f.write(_new)

                    if not force:
                        print("file " + _target + " wrote")

        target_type = Identifier.get_type(target)
        if target_type == IdentifierType.OBI:
            save_dir(target, unit_list)
        elif target_type == IdentifierType.TIO or target_type == IdentifierType.VPL:
            save_file(target, unit_list)
        else:
            print("fail: target " + target + " do not supported for build operation\n")



import unittest
from .compare import compare_text

class Test:
    folder = "tests/03_run/fup"

    def test_run_mixed_side(self, capsys):
        cmd = "-w 80 -m test solver.cpp cases.tio -as"
        compare_text(capsys, Test.folder, "out1", cmd)

    def test_run_side_0(self, capsys):
        cmd = "-w 80 -m test solver.py cases.tio -d"
        compare_text(capsys, Test.folder, "out3", cmd)
        
    def test_run_side_1(self, capsys):
        cmd = "-w 80 -m test solver.py cases.tio -s"
        compare_text(capsys, Test.folder, "out4", cmd)

    def test_run_side_2(self, capsys):
        cmd = "-w 80 -m test solver.py cases.tio -a"
        compare_text(capsys, Test.folder, "out5", cmd)

    def test_run_side_5(self, capsys):
        cmd = "-w 80 -m test solver.cpp cases.tio -ad"
        compare_text(capsys, Test.folder, "out2", cmd)

if __name__ == '__main__':
    unittest.main()



import json
from typing import Any, Dict
from .rep_settings import RepData, RepSource
from .app_settings import AppSettings
import os
import appdirs

from tko.util.text import Text
from tko.play.colors import Colors
from tko.util.raw_terminal import RawTerminal

def singleton(class_):
    instances = {}
    def getinstance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]
    return getinstance

@singleton
class Settings:
    def __init__(self):
        self.reps: Dict[str, RepSource] = {}
        self.app = AppSettings()
        self.colors = Colors()

        self.settings_file = ""

    def set_settings_file(self, path: str):
        self.settings_file = path
        return self

    def get_settings_file(self) -> str:
        if self.settings_file is None or self.settings_file == "":
            self.package_name = "tko"
            default_filename = "settings.json"
            self.settings_file = os.path.abspath(default_filename)  # backup for replit, dont remove
            self.settings_file = os.path.join(appdirs.user_data_dir(self.package_name), default_filename)
        
        if not os.path.exists(self.settings_file):
            os.makedirs(os.path.dirname(self.settings_file), exist_ok=True)
        return self.settings_file

    def reset(self):
        self.reps = {}
        self.reps["fup"] = RepSource(url = "https://github.com/qxcodefup/arcade/blob/master/Readme.md")
        self.reps["ed"] = RepSource(url = "https://github.com/qxcodeed/arcade/blob/master/Readme.md")
        self.reps["poo"] = RepSource(url = "https://github.com/qxcodepoo/arcade/blob/master/Readme.md")

        # for key in self.reps:
        #     repdata = self.get_rep_data(key)
        #     repdata.save_data_to_json()

        self.app = AppSettings()
        self.colors = Colors()
        return self

    def __get_old_rep_file_path(self, course: str) -> str:
        return os.path.join(self.app.get_rootdir(), course, ".rep.json")

    def __get_new_rep_file_path(self, course: str) -> str:
        return os.path.join(self.app.get_rootdir(), course, "rep.json")

    def get_rep_source(self, course: str) -> RepSource:
        if course in self.reps:
            return self.reps[course]
        raise Warning(f"Curso {course} não encontrado")

    def get_rep_data(self, course: str) -> RepData:
        old_cfg_file = self.__get_old_rep_file_path(course)
        new_cfg_file = self.__get_new_rep_file_path(course)
        if os.path.exists(old_cfg_file):
            os.rename(old_cfg_file, new_cfg_file)
        
        rep_data = RepData(self.app.get_rootdir(), course, new_cfg_file)
        if os.path.exists(new_cfg_file):
            return rep_data.load_data_from_json()
        return rep_data.load_defaults()
  
    # def to_dict(self) -> Dict[str, Any]:
    #     return {
    #         "reps": {k: v.to_dict() for k, v in self.reps.items()},
    #         "geral": self.app.to_dict()
    #     }
    
    def load_settings(self):
        try:
            settings_file = self.get_settings_file() # assure right loading if value == ""
            with open(settings_file, "r", encoding="utf-8") as f:
                data = json.load(f)
                self.reps = {k: RepSource().from_dict(v) for k, v in data.get("reps", {}).items()}
                self.app = AppSettings().from_dict(data.get("geral", {}))
                self.colors = Colors().from_dict(data.get("colors", {}))
        except (FileNotFoundError, json.decoder.JSONDecodeError) as _e:
            self.reset()
            self.save_settings()
        return self

    # def from_dict(self, data: Dict[str, Any]):
    #     self.reps = {k: RepSource().from_dict(v) for k, v in data.get("reps", {}).items()}
    #     self.app = AppSettings().from_dict(data.get("geral", {}))
    #     return self

    def check_rootdir(self):
        if self.app._rootdir != "":
            return
        print(Text().add("Pasta padrão para download de arquivos ").addf("r", "precisa").add(" ser definida."))
        here_cwd = os.getcwd()
        qxcode = os.path.join(os.path.expanduser("~"), "qxcode")

        while True:
            print(Text().addf("r", "1").add(" - ").add(here_cwd))
            print(Text().addf("r", "2").add(" - ").add(qxcode))
            print(Text().addf("r", "3").add(" - ").add("Outra pasta"))
            print(Text().add("Default ").addf("r", "1").add(": "), end="")
            op = input()
            if op == "":
                op = "1"
            if op == "1":
                home_qxcode = here_cwd
                break
            if op == "2":
                home_qxcode = qxcode
                break
            if op == "3":
                print(Text().addf("y", "Navegue até o diretório desejado e execute o tko novamente."))
                exit(1)

        if not os.path.exists(home_qxcode):
            os.makedirs(home_qxcode)
        print("Pasta padrão para download de arquivos foi definida em: " + home_qxcode)
        print(RawTerminal.centralize("", "-"))
        self.app._rootdir = home_qxcode
        self.save_settings();
        return self
    
    def check_rep_alias(self, rep_alias: str):
        if rep_alias == "__ask":
            last = self.app.get_last_rep()
            if last != "" and last in self.reps:
                rep_alias = last
            else:
                print("Escolha um dos repositórios para abrir:")
                options: Dict[int, str] = {}
                for i, alias in enumerate(self.reps, start=1):
                    print(Text().addf("r", str(i)).add(f" - {alias}"))
                    options[i] = alias
                while True:
                    try:
                        print("Digite o número do repositório desejado: ", end="")
                        index = int(input())
                        if index in options:
                            rep_alias = options[index]
                            self.app.set_last_rep(rep_alias)
                            self.save_settings()
                            break
                    except ValueError:
                        pass
                    print("Digite um número válido")
        return rep_alias
    
    def save_settings(self):
        file = self.get_settings_file()
        value = {
            "reps": {k: v.to_dict() for k, v in self.reps.items()},
            "geral": self.app.to_dict(),
            "colors": self.colors.to_dict()
        }
        with open(file, "w", encoding="utf-8") as f:
            json.dump(value, f, indent=4)
        return self

    def __str__(self):
        output = ["Repositories:"]
        maxlen = max([len(key) for key in self.reps])
        for key in self.reps:
            prefix = f"- {key.ljust(maxlen)}"
            if self.reps[key].file and self.reps[key].url:
                output.append(f"{prefix} : dual   : {self.reps[key].url} ; {self.reps[key].file}")
            elif self.reps[key].url:
                output.append(f"{prefix} : remote : {self.reps[key].url}")
            else:
                output.append(f"{prefix} : local  : {self.reps[key].file}")
        return "\n".join(output)



from typing import List, Dict, Tuple, Optional
from .quest import Quest
from .game import Game
import pydot # type: ignore


class Graph:
    graph_generated: bool = False

    colorlist: List[Tuple[str, str]] = [
            ("aquamarine3", "aquamarine4"),
            ("bisque3", "bisque4"),
            ("brown3", "brown4"),
            ("chartreuse3", "chartreuse4"),
            ("coral3", "coral4"),
            ("cyan3", "cyan4"),
            ("darkgoldenrod3", "darkgoldenrod4"),
            ("darkolivegreen3", "darkolivegreen4"),
            ("darkorchid3", "darkorchid4"),
            ("darkseagreen3", "darkseagreen4"),
            ("darkslategray3", "darkslategray4"),
            ("deeppink3", "deeppink4"),
            ("deepskyblue3", "deepskyblue4"),
            ("dodgerblue3", "dodgerblue4"),
            ("firebrick3", "firebrick4"),
            ("gold3", "gold4"),
            ("green3", "green4"),
            ("hotpink3", "hotpink4"),
            ("indianred3", "indianred4"),
            ("khaki3", "khaki4"),
            ("lightblue3", "lightblue4"),
            ("lightcoral", "lightcoral"),
            ("lightcyan3", "lightcyan4"),
            ("lightgoldenrod3", "lightgoldenrod4"),
            ("lightgreen", "lightgreen"),
            ("lightpink3", "lightpink4"),
            ("lightsalmon3", "lightsalmon4"),
            ("lightseagreen", "lightseagreen"),
            ("lightskyblue3", "lightskyblue4"),
            ("lightsteelblue3", "lightsteelblue4"),
            ("lightyellow3", "lightyellow4"),
            ("magenta3", "magenta4"),
            ("maroon3", "maroon4"),
            ("mediumorchid3", "mediumorchid4"),
            ("mediumpurple3", "mediumpurple4"),
            ("mediumspringgreen", "mediumspringgreen"),
            ("mediumturquoise", "mediumturquoise"),
            ("mediumvioletred", "mediumvioletred"),
            ("mistyrose3", "mistyrose4"),
            ("navajowhite3", "navajowhite4"),
            ("olivedrab3", "olivedrab4"),
            ("orange3", "orange4"),
            ("orangered3", "orangered4"),
            ("orchid3", "orchid4"),
            ("palegreen3", "palegreen4"),
            ("paleturquoise3", "paleturquoise4"),
            ("palevioletred3", "palevioletred4")
            ]

    def __init__(self, game: Game):
        self.game = game
        self.path = "graph.png"
        self.opt = False
        self.reachable = [q.key for q in self.game.quests.values() if q.is_reachable()]
        # self.reachable = [key for key in self.game.quests.keys()]

        self.counts: Dict[str, str] = {}
        for q in self.game.quests.values():
            done = len([t for t in q.get_tasks() if t.is_complete()])
            init = len([t for t in q.get_tasks() if t.in_progress()])
            todo = len([t for t in q.get_tasks() if t.not_started()])
            self.counts[q.key] = f"{done} / {done + init + todo}"

    def set_path(self, path: str):
        self.path = path
        return self

    def set_opt(self, opt: bool):
        self.opt = opt
        return self

    def set_counts(self, counts: Dict[str, str]):
        self.counts = counts
        return self
    
    def set_output(self, output: str):
        self.output = output
        return self

    def info(self, qx: Quest):
        text = f'{qx.title.strip()}'
        if self.reachable is None or self.counts is None:
            return f'"{text}"'
        return f'"{text}\\n{self.counts[qx.key]}"'

    def is_reachable_or_next(self, q: Quest):
        if self.reachable is None:
            return True
        if q.key in self.reachable:
            return True
        for r in q.requires_ptr:
            if r.key in self.reachable:
                return True
        return False

    def generate(self):
        saida = ["digraph diag {", '  node [penwidth=1, style="rounded,filled", shape=box]']

        targets = [q for q in self.game.quests.values()]
        for q in targets:
            token = "->"
            if len(q.requires_ptr) > 0:
                for r in q.requires_ptr:
                    extra = ""
                    if self.reachable is not None:
                        if q.key not in self.reachable and not r.is_complete():
                            extra = "[style=dotted]"
                    saida.append(f"  {self.info(r)} {token} {self.info(q)} {extra}")
            else:
                v = '  "Início"'
                saida.append(f"{v} {token} {self.info(q)}")

        for i, c in enumerate(self.game.clusters.values()):
            # cluster_targets = [q for q in c.quests if self.is_reachable_or_next(q)]
            cluster_targets = [q for q in c.quests]
            for q in cluster_targets:
                if self.opt:
                    if q.opt:
                        fillcolor = "pink"
                    else:
                        fillcolor = "lime"
                else:
                    if c.color is not None:
                        fillcolor = c.color
                    else:
                        fillcolor = self.colorlist[i][0]

                    if q.opt:
                        fillcolor = f'"{fillcolor};0.9:orange"'
                    else:
                        fillcolor = f'"{fillcolor};0.9:lime"'
                shape = "ellipse"
                color = "black"
                width = 1
                if self.reachable is not None:
                    if q.key not in self.reachable:
                        color = "white"
                    else:
                        width = 3
                        color = q.get_grade_color()
                        if color == "g":
                            color = "green"
                        elif color == "r":
                            color = "red"
                        elif color == "y":
                            color = "yellow"
                        elif color == "m":
                            color = "magenta"
                saida.append(f"  {self.info(q)} [shape={shape} color={color} penwidth={width} fillcolor={fillcolor} ]")

        saida.append("}")
        # saida.append("@enduml")
        saida.append("")

        # dot_file = os.path.join(os.path.dirname(self.path) + "graph.dot")
        # open(dot_file, "w").write("\n".join(saida))
        out_file = self.path
        data = "\n".join(saida)
        graph_dot  = pydot.graph_from_dot_data(data)[0]
        graph_dot.write_png(out_file)



from typing import Optional, List, Callable
import os
import urllib.request
import urllib.error
import json

from tko.down.down import DownProblem

from ..settings.settings import Settings
from ..game.game import Game
from ..util.remote import RemoteCfg


class CmdDown:
    @staticmethod
    def execute(rep_alias: str, task_key: str, language: Optional[str], settings: Settings, fnprint: Callable[[str], None], game: Optional[Game] = None) -> bool:
        DownProblem.fnprint = fnprint
        rep_dir = os.path.join(settings.app.get_rootdir(), rep_alias)
        rep_source = settings.get_rep_source(rep_alias)
        rep_data = settings.get_rep_data(rep_alias)
        if game is None:
            try:
                file = rep_source.get_file_or_cache(rep_dir)
            except urllib.error.HTTPError:
                DownProblem.fnprint("falha: Verifique sua internet")
            game = Game(file)
        item = game.get_task(task_key)
        if not item.link.startswith("http"):
            DownProblem.fnprint("falha: link para atividade não é um link remoto")
            return False
        cfg = RemoteCfg(item.link)
        cache_url = os.path.dirname(cfg.get_raw_url()) + "/.cache/"

        destiny = DownProblem.create_problem_folder(rep_dir, task_key)
        destiny = os.path.abspath(destiny)
        try:
            [_readme_path, mapi_path] = DownProblem.down_problem_def(destiny, cache_url)
        except urllib.error.HTTPError:
            DownProblem.fnprint("  falha: atividade não encontrada no curso")
            # verifi if destiny folder is empty and remove it
            if len(os.listdir(destiny)) == 0:
                os.rmdir(destiny)
            return False
        except urllib.error.URLError:
            DownProblem.fnprint("  falha: não consegui baixar a atividade, verifique sua internet")
            return False


        with open(mapi_path, encoding="utf-8") as f:
            loaded_json = json.load(f)
        os.remove(mapi_path)

        language_def = rep_data.get_lang()
        if language_def == "":
            language_def = Settings().app.get_lang_default()
        ask_ext = False
        if language is None:
            if language_def != "":
                language = language_def
            else:
                print("  Escolha uma extensão para os rascunhos: [c, cpp, py, ts, js, java]: ", end="")
                language = input()
                ask_ext = True

        DownProblem.unpack_json(loaded_json, destiny, language)
        DownProblem.download_drafts(loaded_json, destiny, language, cache_url, ask_ext)
        return True



import unittest

from .compare import compare_text

class Test:
    def test_all(self, capsys):
        f = "tests/06_failures"
        compare_text(capsys, f, "out1", "-w 80 -m test cases.tio draft.ts -ad")
        compare_text(capsys, f, "out2", "-w 80 -m test cases.tio draft.ts -d")
        compare_text(capsys, f, "out3", "-w 80 -m test cases.tio draft.ts -qd")

if __name__ == '__main__':
    unittest.main()



import unittest
import os

from .compare import compare_list, compare_text


class Test:
    folder = "tests/03_run/sql"

    def test_run_1(self, capsys):
        cmd_run = "cat create.sql solver.sql | sqlite3"
        cmd = ["-m", "test", "--cmd", cmd_run, "cases.tio"]
        compare_list(capsys, Test.folder, "out1", cmd)


    def test_run_2(self, capsys):
        compare_text(capsys, Test.folder, "out2", "-w 80 -m test create.sql student.sql")
    

if __name__ == '__main__':
    unittest.main()



from .unit import Unit
from ..util.consts import ExecutionResult

from ..util.runner import Runner
from .solver_builder import SolverBuilder
from typing import Optional

class UnitRunner:

    def __init__(self):
        pass

    # run a unit using a solver and return if the result is correct
    @staticmethod
    def run_unit(solver: SolverBuilder, unit: Unit, timeout: Optional[float]=None) -> ExecutionResult:
        if solver.compile_error:
            unit.received = solver.error_msg
            return ExecutionResult.COMPILATION_ERROR
        cmd = solver.get_executable()
        if timeout == 0:
            timeout = None
        return_code, stdout, stderr = Runner.subprocess_run(cmd, unit.input, timeout)
        unit.received = stdout + stderr
        if return_code != 0:
            return ExecutionResult.EXECUTION_ERROR
        if unit.received == unit.expected:
            return ExecutionResult.SUCCESS
        return ExecutionResult.WRONG_OUTPUT


from urllib.request import urlopen
from typing import Optional
from .. import __version__

class CheckVersion:

    link = "https://raw.githubusercontent.com/senapk/tko/master/src/tko/__init__.py"

    def __init__(self):
        self.version: str = __version__
        self.latest_version: Optional[str] = self.get_latest_version()

    def version_check(self):
        if self.latest_version is None:
            return
        if self.version != self.latest_version:
            major, minor, patch = [int(x) for x in self.version.split(".")]
            latest_major, latest_minor, latest_patch = [int(x) for x in self.latest_version.split(".")]
            if major < latest_major or (major == latest_major and minor < latest_minor) or (major == latest_major and minor == latest_minor and patch < latest_patch):
                print(f"Sua versão do  TKO ({self.version}) está desatualizada.")
                print(f"A última versão é a {self.latest_version}.")

    def get_latest_version(self):
        try:
            with urlopen(self.link) as f:
                for line in f:
                    if b"__version__" in line:
                        return line.decode().split('"')[1]
        except:
            return None